import os, sys, copy, re

import libtbx.phil

import numpy
import pandas

from giant.xray.data import CrystalSummary
from giant.xray.data.cluster import CrystalGroup

#######################################

blank_arg_prepend = {'.mtz':'mtz=', '.pdb':'pdb='}

master_phil = libtbx.phil.parse("""
input {
    mtz = None
        .type = path
        .multiple = True
    mtz_regex = None
        .type = str

    pdb = None
        .type = path
        .multiple = True
    pdb_regex = None
        .type = str
}
clustering{
    lcv_cutoff = 0.2
        .type = float
    label_nodes_above = 0.2
        .help = "Label nodes above this value"
        .type = float
}
output {
    out_dir = None
        .type = path
    split_pdbs_and_mtzs = True
        .type = bool
    html_out = None
        .type = path
}
""")

#######################################

bar = '============================================>'

#######################################

def run(params):

    if not (params.input.pdb or params.input.mtz): raise Exception('No pdb/mtz files have been provided.')

    # Make output directory if required
    if params.output.out_dir and (not os.path.exists(params.output.out_dir)):
        os.mkdir(params.output.out_dir)
    # Dump images into the current directory if no directory is given
    if not params.output.out_dir: img_dir = '.'
    else: img_dir = params.output.out_dir

    #########################################################################################################

    print bar
    print 'Making pdb/mtz dataset labels'

    try:
        if params.input.pdb_regex: p_labels = [re.findall(params.input.pdb_regex, f)[0] for f in params.input.pdb]
        else:                      p_labels = ['PDB-{:06d}'.format(i) for i in range(len(params.input.pdb))]
        if params.input.mtz_regex: m_labels = [re.findall(params.input.mtz_regex, f)[0] for f in params.input.mtz]
        else:                      m_labels = ['MTZ-{:06d}'.format(i) for i in range(len(params.input.mtz))]
    except:
        print 'Error parsing: {}'.format(f)
        raise

    #########################################################################################################

    print bar
    print 'Reading {} pdb(s) and {} mtz(s)'.format(len(params.input.pdb), len(params.input.mtz))

    if params.input.pdb: pdb_summaries = [CrystalSummary.from_pdb(pdb_file=f, id=lab) for f,lab in zip(params.input.pdb, p_labels)]
    else:                pdb_summaries = []
    if params.input.mtz: mtz_summaries = [CrystalSummary.from_mtz(mtz_file=f, id=lab) for f,lab in zip(params.input.mtz, m_labels)]
    else:                mtz_summaries = []

    print bar
    print 'Grouping crystals by space group...'

    # Group by SpaceGroup
    crystal_groups = CrystalGroup.by_space_group(crystals=pdb_summaries+mtz_summaries)

    print '...grouped crystals into {} groups'.format(len(crystal_groups))

    #########################################################################################################

    print bar
    print 'Analysing variation of unit cells for each space group'

    for cg in crystal_groups:

        sg_name = 'SG-{}'.format(cg.space_groups[0].replace(' ','').replace('(','-').replace(')',''))

        print bar
        print ''
        print bar
        print 'Space Group {}: {} dataset(s)'.format(cg.space_groups[0], len(cg.crystals))

        #########################################################################################################

        print '===========================>'
        print 'Unit Cell Variation:'
        print numpy.round(cg.uc_stats.as_pandas_table().T, 2)

        #########################################################################################################

        print '===========================>'
        print 'Making unit cell dendrogram for all crystals with this spacegroup'
        if len(cg.crystals) > 1:
            cg.dendrogram(  fname = os.path.join(img_dir,'{}-dendrogram.png'.format(sg_name)),
                            xlab  = 'Crystal',
                            ylab  = 'Linear Cell Variation')

        #########################################################################################################

        print '===========================>'
        print 'Clustering unit cells...'

        sg_crystal_groups = cg.by_unit_cell(cg.crystals, cutoff=params.clustering.lcv_cutoff)

        print '...clustered crystals into {} groups'.format(len(sg_crystal_groups))

        for i_cg2, cg2 in enumerate(sg_crystal_groups):

            cluster_name = '{}-Cluster-{}'.format(sg_name, i_cg2+1)

            print '===========================>'
            print 'Processing: {}'.format(cluster_name)

            #########################################################################################################

            print 'Making unit cell dendrogram for this cluster of crystals'

            if len(cg2.crystals) > 1:
                cg2.dendrogram( fname = os.path.join(img_dir, '{}-dendrogram.png'.format(cluster_name)),
                                xlab  = 'Crystal',
                                ylab  = 'Linear Cell Variation',
                                ylim  = (0, params.clustering.lcv_cutoff),
                                annotate_y_min = params.clustering.label_nodes_above)

            #########################################################################################################

            if params.output.out_dir:
                print 'Making and populating output directory'

                # Go through and link the datasets for each of the spacegroups into a separate folder
                sub_dir = os.path.join(params.output.out_dir, cluster_name)
                if not os.path.exists(sub_dir): os.mkdir(sub_dir)

                if params.output.split_pdbs_and_mtzs:
                    # Split the mtzs and pdbs into separate directories
                    mtz_dir = os.path.join(sub_dir, 'mtzs')
                    if not os.path.exists(mtz_dir): os.mkdir(mtz_dir)
                    pdb_dir = os.path.join(sub_dir, 'pdbs')
                    if not os.path.exists(pdb_dir): os.mkdir(pdb_dir)
                else:
                    mtz_dir = pdb_dir = sub_dir

                for c in cg2.crystals:
                    if c.mtz_file:
                        # Create subdirectory
                        sub_sub_dir = os.path.join(mtz_dir, c.id)
                        if not os.path.exists(sub_sub_dir): os.mkdir(sub_sub_dir)
                        os.symlink(os.path.abspath(c.mtz_file), os.path.join(sub_sub_dir, c.id+'.mtz'))
                        # Link PDB as well if filenames are the same
                        potential = c.mtz_file.replace('.mtz','.pdb')
                        if os.path.exists(potential):
                            os.symlink(os.path.abspath(potential), os.path.join(sub_sub_dir, c.id+'.pdb'))
                    if c.pdb_file:
                        # Create subdirectory
                        sub_sub_dir = os.path.join(pdb_dir, c.id)
                        if not os.path.exists(sub_sub_dir): os.mkdir(sub_sub_dir)
                        os.symlink(os.path.abspath(c.pdb_file), os.path.join(sub_sub_dir, c.id+'.pdb'))
                        # Link PDB as well if filenames are the same
                        potential = c.pdb_file.replace('.pdb','.mtz')
                        if os.path.exists(potential):
                            os.symlink(os.path.abspath(potential), os.path.join(sub_sub_dir, c.id+'.mtz'))

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



from scitbx.math import basic_statistics

from giant.structure.select import non_h, protein, backbone, sidechains

########################################################################################

class BfactorStatistics(object):
    """B-factor statstics for a structure"""
    def __init__(self, all=None, protein=None, backbone=None, sidechain=None):
        self.all       = all
        self.protein   = protein
        self.backbone  = backbone
        self.sidechain = sidechain

    def to_z_score(self, b_vals, method='backbone'):
        """Convert a set of B-factors to Z-scores"""
        assert method in ['all','protein','backbone','sidechain']
        if method == 'all':       stats = self.all
        if method == 'protein':   stats = self.protein
        if method == 'backbone':  stats = self.backbone
        if method == 'sidechain': stats = self.sidechain
        if stats.biased_standard_deviation == 0.0:
            # No variation - return list of zeroes
            return b_vals-b_vals

        return (b_vals - stats.mean)/stats.biased_standard_deviation

    @classmethod
    def from_pdb(cls, pdb_input=None, pdb_hierarchy=None):
        """Calculate the b-factor statistics of a model"""

        assert [pdb_input, pdb_hierarchy].count(None)==1,'Provide pdb_input OR pdb_hierarchy'
        if pdb_input: pdb_hierarchy = pdb_input.construct_hierarchy()

        cache = pdb_hierarchy.atom_selection_cache()

        all_b       = non_h(hierarchy=pdb_hierarchy, cache=cache, copy=True).atoms().extract_b()
        protein_b   = protein(hierarchy=pdb_hierarchy, cache=cache, copy=True).atoms().extract_b()
        backbone_b  = backbone(hierarchy=pdb_hierarchy, cache=cache, copy=True).atoms().extract_b()
        sidechain_b = sidechains(hierarchy=pdb_hierarchy, cache=cache, copy=True).atoms().extract_b()

        return cls(all       = basic_statistics(all_b),
                   protein   = basic_statistics(protein_b),
                   backbone  = basic_statistics(backbone_b),
                   sidechain = basic_statistics(sidechain_b) )

    def show(self):
        if self.all:
            print '================>'
            print 'All Atoms:'
            print self.all.show()
        if self.protein:
            print '================>'
            print 'Protein Atoms:'
            print self.protein.show()
        if self.backbone:
            print '================>'
            print 'Backbone Atoms:'
            print self.backbone.show()
        if self.sidechain:
            print '================>'
            print 'Sidechain Atoms:'
            print self.sidechain.show()

########################################################################################

def normalise_b_factors_to_z_scores(pdb_input=None, pdb_hierarchy=None, b_factor_statistics=None, method='backbone'):
    """Calculate the b-factor statistics of the model"""

    assert [pdb_input, pdb_hierarchy].count(None)==1,'Provide pdb_input OR pdb_hierarchy'
    if pdb_input: pdb_hierarchy = pdb_input.construct_hierarchy()

    if not b_factor_statistics: b_factor_statistics = BfactorStatistics.from_pdb(pdb_hierarchy=pdb_hierarchy)
    new_b = b_factor_statistics.to_z_score(b_vals=pdb_hierarchy.atoms().extract_b(), method=method)

    output_h = pdb_hierarchy.deep_copy()
    output_h.atoms().set_b(new_b)

    return output_h

########################################################################################





from bamboo.macro.utils import get_residue_labels

def extract_scores(scores, label):
    """Pull out the score specified by 'label'"""

    # Validate label (NAME-CHAIN-NUMBER-INSCODE)
    if label[3] == '':
        label = list(label)
        label[3] = ' '
        label = tuple(label)

    # Extract Scores
    try:
        labscores = scores[label]
    except KeyError:
        if label[1] == ' ':
            reflabel=list(label)
            reflabel[1] = '_'
            if tuple(reflabel) in scores:
                labscores = scores[tuple(reflabel)]
            else:
                reflabel[1] = 'X'
                labscores = scores[tuple(reflabel)]
        else:
            print 'Scores not found in EDSTATS output!'
            raise

    return labscores

def score_against_density_by_subset_file(mtz_file, pdb_file, subpdb):
    """Scores all the residues with edstats and returns the scores for the residues present in `subpdb`"""

    # Get the labels
    labels = get_residue_labels(subpdb)
    # Score by labels
    labels, summary, scores = score_against_density_by_res_labels(mtz_file, pdb_file, labels)

    return labels, summary, scores

def score_against_density_by_res_labels(mtz_file, pdb_file, labels):
    """Scores all the residues with edstats and returns the scores for the residues identified by `labels`"""

    # Get the scores
    edstats, summary = score_file_with_edstats(mtz_file, pdb_file)
    # Score by label (list of tuples)
    scores = dict([(label,extract_scores(edstats.scores, label)) for label in labels])
    # Add summary percentiles for the labelled residue
    summary.percentiles = dict([(label,get_labelled_residue_percentiles(label, edstats.scores, summary)) for label in labels])

    return labels, summary, scores

def get_labelled_residue_percentiles(label, edstats_scores):
    """Takes the scores and summary object from edstats and compares the residue given by 'label' to the rest of the protein"""

    # Get the res scores given by the label
    res_score = extract_scores(edstats_scores, label)

    # Get the total number of residues in the file
    num_res = edstats_scores.columns.size

    # Get the list of ZD scores
    all_ZD =       edstats_scores.loc['ZDa']
    # Get the list of ZD+ scores
    all_ZD_plus =  edstats_scores.loc['ZD+a']
    # Get the list of ZD- scores
    all_ZD_minus = edstats_scores.loc['ZD-a']
    # Get the list of ZO scores
    all_ZO =       edstats_scores.loc['ZOa']
    # Get the list of RSR Scores
    all_RSR =      edstats_scores.loc['Ra']
    # Get the list of RSCC Scores
    all_RSCC =     edstats_scores.loc['CCPa']
    # Get the list of B-factor Scores
    all_B =        edstats_scores.loc['BAa']

    # Get the percentiles of the ligand scores
    percentiles = {}
    percentiles['%ile_ZD']  =  (len([i for i in all_ZD       if i < res_score['ZDa']])      + len([i for i in all_ZD       if i <= res_score['ZDa']])) * 50 / float(num_res)
    percentiles['%ile_ZD+']  = (len([i for i in all_ZD_plus  if i < res_score['ZD+a']])     + len([i for i in all_ZD_plus  if i <= res_score['ZD+a']])) * 50 / float(num_res)
    percentiles['%ile_ZD-']  = (len([i for i in all_ZD_minus if i > res_score['ZD-a']])     + len([i for i in all_ZD_minus if i >= res_score['ZD-a']])) * 50 / float(num_res)
    percentiles['%ile_ZO']   = (len([i for i in all_ZO       if i > res_score['ZOa']])      + len([i for i in all_ZO       if i >= res_score['ZOa']])) * 50 / float(num_res)
    percentiles['%ile_RSR']  = (len([i for i in all_RSR      if i < res_score['Ra']])       + len([i for i in all_RSR      if i <= res_score['Ra']])) * 50 / float(num_res)
    percentiles['%ile_RSCC'] = (len([i for i in all_RSCC     if i > res_score['CCPa']])     + len([i for i in all_RSCC     if i >= res_score['CCPa']])) * 50 / float(num_res)
    percentiles['%ile_B']    = (len([i for i in all_B        if i < res_score['BAa']])      + len([i for i in all_B        if i <= res_score['BAa']])) * 50 / float(num_res)

    percentiles['%ile_ZOs']   = (len([i for i in all_ZO       if i > res_score['ZOs']])      + len([i for i in all_ZO       if i >= res_score['ZOs']])) * 50 / float(num_res)
    percentiles['%ile_RSCCs'] = (len([i for i in all_RSCC     if i > res_score['CCPs']])     + len([i for i in all_RSCC     if i >= res_score['CCPs']])) * 50 / float(num_res)
    percentiles['%ile_Bs']    = (len([i for i in all_B        if i < res_score['BAs']])      + len([i for i in all_B        if i <= res_score['BAs']])) * 50 / float(num_res)

    return percentiles




import os, sys
import pkg_resources

PANDDA_LIB_TOP = os.path.realpath(os.path.abspath(os.path.join(__path__[0], '..')))
try:
    PANDDA_VERSION = pkg_resources.get_distribution("panddas").version
except:
    PANDDA_VERSION = 'xxx'

PANDDA_TEXT = """
------------------------------------------------------------------>
-
-  Pandda Version {!s}
-
-  __________                    .___  .___
-  \\______   \\_____    ____    __| _/__| _/____    ______
-   |     ___/\\__  \\  /    \\  / __ |/ __ |\\__  \\  /  ___/
-   |    |     / __ \\|   |  \\/ /_/ / /_/ | / __ \\_\\___ \\
-   |____|    (____  /___|  /\\____ \\____ |(____  /____  >
-                  \\/     \\/      \\/    \\/     \\/     \\/
-
-  PAN-Dataset Density Analysis of Crystallographic Fragment Screens
-
-  When using this software, please cite:
-
-       Unpublished,
-       N Pearce, et al.
-
-  and include "PANDDA" in the KEYWRD section of related PDB files
-
------------------------------------------------------------------>
""".format(PANDDA_VERSION)

def welcome(user=None):
    """Welcome message"""

    if user is None:
        try:    user = os.getlogin()
        except: user = "Unknown user"

    try:
        from greetings import get_greeting
        print get_greeting(username=user)
    except SystemExit: raise
    except: pass

    try:
        print 'Hi {!s}. Welcome to Pandda.'.format(user.upper())
    except: pass




import os, sys
import math

from bamboo.rdkit_utils.coords.utils import get_centroid_from_file, calculate_coordinate_differences
from bamboo.wrappers.symutils import map_to_reference_using_symmetry

from rdkit import Chem

def calculate_fraction_of_atom_pairs_within_tolerance(model1, model2, tolerance=1, suppress=False):
    """Calculate the percentage of atoms in model1 that are within `tolerance` of the corresponding atom in model2"""

    try: differences = calculate_coordinate_differences(model1, model2)
    except EqualityError as err:
        if suppress:
            print '{!s}: {!s}'.format(type(err),str(err))
            return None
        else: raise
    # Check for failures
    if not differences:
        if suppress: return None
        else:        raise Exception('Error calculating coordinate differences.\n\t{!s}\n\t{!s}'.format(model1,model2))
    # Calculate the lengths and whether that length is within tolerance
    within_tolerance_all = []
    for diffset in differences:
        within = sum([1.0 for d in diffset if (d[0]**2 + d[1]**2 + d[2]**2 < tolerance**2)])
        within_tolerance_all.append(within/len(diffset))
    # Return the rmsd corresponding to the minimal matching pattern
    return max(within_tolerance_all)

def calculate_fraction_of_atom_pairs_within_tolerance_using_symmetry(model1, model2, tolerance=1, suppress=False, delete_structures=True):
    """Calculate the percentage of atoms in model1 that are within `tolerance` of the corresponding atom in model2, maximising over equivalent models"""

    # Map model2 to model1 using symmetry
    tempfile = map_to_reference_using_symmetry(refpdb=model1, movpdb=model2, pdbout=model2.replace('.pdb','.temp'))
    # Calculate the RMSD as normal
    percentage = calculate_fraction_of_atom_pairs_within_tolerance(model1, tempfile, tolerance, suppress)
    # Remove tempfiles and return
    if delete_structures: os.remove(tempfile)
    return percentage

def calculate_rmsd(model1, model2, suppress=False):
    """Calculates the standard Root Mean Squared Deviation between the coordinates of the models"""

    try: differences = calculate_coordinate_differences(model1, model2)
    except EqualityError as err:
        if suppress:
            print '{!s}: {!s}'.format(type(err),str(err))
            return None
        else: raise
    # Check for failures
    if not differences:
        if suppress: return None
        else:        raise Exception('Error calculating coordinate differences.\n\t{!s}\n\t{!s}'.format(model1,model2))
    # Calculate the rms of the differences
    RMSDs = []
    for diffset in differences:
        sqrsum = sum([d[0]**2 + d[1]**2 + d[2]**2 for d in diffset])
        RMSDs.append(math.sqrt(sqrsum/len(diffset)))
    # Return the rmsd corresponding to the minimal matching pattern
    return min(RMSDs)

def calculate_rmsd_using_symmetry(model1, model2, suppress=False, delete_structures=True):
    """Calculates the standard Root Mean Squared Deviation between the coordinates of the models, minimising over equivalent models"""

    # Map model2 to model1 using symmetry
    tempfile = map_to_reference_using_symmetry(refpdb=model1, movpdb=model2, pdbout=model2.replace('.pdb','.temp'))
    # Calculate the RMSD as normal
    RMSD = calculate_rmsd(model1, tempfile, suppress)
    # Remove tempfiles and return
    if delete_structures: os.remove(tempfile)
    return RMSD

def calculate_centroid_difference(model1, model2):
    """Calculates the length of the vector connecting the centroids of the two models"""

    centroid1 = get_centroid_from_file(model1)
    centroid2 = get_centroid_from_file(model2)
    diff_vect = [a-b for a,b in zip(centroid1,centroid2)]
    len_diff_vect = math.sqrt(sum([c**2 for c in diff_vect]))
    return len_diff_vect

def calculate_centroid_difference_using_symmetry(model1, model2, delete_structures=True):
    """Calculates the length of the vector connecting the centroids of the two models, minimising over equivalent models"""

    # Map model2 to model1 using symmetry
    tempfile = map_to_reference_using_symmetry(refpdb=model1, movpdb=model2, pdbout=model2.replace('.pdb','.temp'))
    # Calculate the centroid as normal
    cent = calculate_centroid_difference(model1, tempfile)
    # Remove tempfiles and return
    if delete_structures: os.remove(tempfile)
    return cent




import numpy

def scale_map_to_reference(ref_map, map, mask_idxs):
    """Scale given map to reference over the point_mask"""

    # Select values for the masked points
    ref_vals = ref_map.select(mask_idxs)
    map_vals =     map.select(mask_idxs)

    # Fit ref = a*map + b
    # a, b
    scale, offset = numpy.polyfit(x=map_vals, y=ref_vals, deg=1)
#    print '\nScale: {}\tOffset: {}'.format(scale, offset)

    return (map*scale) + offset



import os, sys, shutil

from bamboo.common.command import CommandManager
from bamboo.macro.utils import get_residue_labels
from bamboo.constants import DEFAULT_OUTPUT_CHAIN, DEFAULT_OUTPUT_RESNUM

def isolate_residue(inpdb, outpdb, resname):
    """Extract the residues identified by resname using pdbcur - i.e. 'UNL'"""

    PDBCUR = CommandManager('pdbcur')
    PDBCUR.add_command_line_arguments('XYZIN',inpdb,'XYZOUT',outpdb)
    PDBCUR.add_standard_input(['lvresidue /*/*/({!s})'.format(resname),'END'])
    PDBCUR.run()

    return PDBCUR

def remove_residue(inpdb, outpdb, resname):
    """Delete the residues identified by resname using pdbcur - i.e. 'UNL'"""

    PDBCUR = CommandManager('pdbcur')
    PDBCUR.add_command_line_arguments('XYZIN',inpdb,'XYZOUT',outpdb)
    PDBCUR.add_standard_input(['delresidue /*/*/({!s})'.format(resname),'END'])
    PDBCUR.run()

    return PDBCUR

def change_residue_chain_and_number_to_default(inpdb):
    """Takes the residue in inpdb and moves it to chain X, residue 666 - IN PLACE (keeps original copy as inpdb+'.origfile')"""

    # Get the current labels
    res_id = get_residue_labels(inpdb)
    assert len(res_id)!=0, 'NO RESIDUES IN FILE! {!s}'.format(inpdb)
    assert len(res_id)==1, 'MORE THAN ONE RESIDUE IS PRESENT IN THE FILE! {!s}'.format(inpdb)
    curr_resname, curr_chain, curr_resnum, curr_inscode = res_id[0]
    if not curr_chain: curr_chain='\'\''
    new_chain, new_resnum = DEFAULT_OUTPUT_CHAIN[0], DEFAULT_OUTPUT_RESNUM[0]

    # Move file to prepare for new file
    origfile = inpdb + '.origfile'
    shutil.move(inpdb, origfile)

    PDBSET = CommandManager('pdbset')
    PDBSET.add_command_line_arguments('XYZIN',origfile,'XYZOUT',inpdb)
    PDBSET.add_standard_input(['RENUMBER {!s} CHAIN {!s} TO {!s}'.format(new_resnum, curr_chain, new_chain), 'END'])
    PDBSET.run()

    return PDBSET

def isolate_residue_by_res_id(inpdb, outpdb, chain, resnum, model='*', inscode=''):
    """Isolate the residues identified by residue ids using pdbcur - i.e. '0/A/54.A/'"""

    if inscode:
        selection = '/{!s}/{!s}/{!s}.{!s}'.format(model,chain,resnum,inscode)
    else:
        selection = '/{!s}/{!s}/{!s}'.format(model,chain,resnum)

    PDBCUR = CommandManager('pdbcur')
    PDBCUR.add_command_line_arguments('XYZIN',inpdb,'XYZOUT',outpdb)
    PDBCUR.add_standard_input(['lvresidue {!s}'.format(selection),'END'])
    PDBCUR.run()

    return PDBCUR

def remove_residue_by_res_id(inpdb, outpdb, chain, resnum, model='*', inscode='', removeSolvent=False):
    """Remove the residues identified by res info using pdbcur - i.e. 'UNL'"""

    if inscode:
        selection = '/{!s}/{!s}/{!s}.{!s}'.format(model,chain,resnum,inscode)
    else:
        selection = '/{!s}/{!s}/{!s}'.format(model,chain,resnum)

    std_input = ['delresidue {!s}'.format(selection)]+(removeSolvent)*['delsolvent']+['END']

    PDBCUR = CommandManager('pdbcur')
    PDBCUR.add_command_line_arguments('XYZIN',inpdb,'XYZOUT',outpdb)
    PDBCUR.add_standard_input(std_input)
    PDBCUR.run()

    return PDBCUR

def merge_pdb_files(pdb1, pdb2, pdbout):
    """Merge two PDB Files using CCP4s pdb_merge"""

    # Initialise Commander
    MERGER = CommandManager('pdb_merge')
    # Set command arguments
    MERGER.add_command_line_arguments('xyzin1', pdb1, 'xyzin2', pdb2, 'xyzout', pdbout)
    # Set inputs
    MERGER.add_standard_input('END')
    # run!
    MERGER.run()

    return MERGER

def reset_pdb_file(pdbin, pdbout):
    """Resets the B-Factors in a file and removes anisotropy"""

    if not pdbout.endswith('.pdb'):
        pdbout = pdbout+'.pdb'
    if not os.path.exists(pdbin):
        raise IOError('pdbin does not exist! {!s}'.format(pdbin))
    if os.path.exists(pdbout):
        raise Exception('pdbout already exists! {!s}'.format(pdbout))

    pdbtemp = pdbout.replace('.pdb','.temp.pdb')

    # Initialise Commander
    PDBCUR = CommandManager('pdbcur')
    # Set Command Arguments
    PDBCUR.add_command_line_arguments('XYZIN',pdbin,'XYZOUT',pdbtemp)
    # Set inputs
    PDBCUR.add_standard_input(['NOANISOU','DELSOLVENT','END'])
    # run!
    PDBCUR.run()

    if not os.path.exists(pdbtemp):
        raise ExternalProgramError('PDBCUR has failed to remove anisotropy and delete solvent. {!s}\nOUT: {!s}\nERR: {!s}'.format(pdbin, PDBCUR.out, PDBCUR.err))

    # Initialise Commander
    PDBSET = CommandManager('pdbset')
    # Set Command Arguments
    PDBSET.add_command_line_arguments('XYZIN',pdbtemp,'XYZOUT',pdbout)
    # Set inputs
    PDBSET.add_standard_input(['BFACTOR','END'])
    # run!
    PDBSET.run()

    if not os.path.exists(pdbout):
        raise ExternalProgramError('PDBSET has failed to reset B-factors. {!s}\nOUT: {!s}\nERR: {!s}'.format(pdbtemp, PDBSET.out, PDBSET.err))
    else:
        os.remove(pdbtemp)

    return pdbout

def remove_modres_records(pdb):
    """greps out lines beginning with MODRES from a file"""

    assert os.path.exists(pdb), 'input file must exist'

    origpdb = pdb + '.origfile'
    assert not os.path.exists(origpdb), 'temporary file must not exist'

    shutil.move(pdb, origpdb)

    cmd = "grep -v '^MODRES' {!s} > {!s}".format(origpdb, pdb)
    os.system(cmd)

    if not os.path.exists(pdb):
        raise ExternalProgramError('Failed to remove modres records from {!s}'.format(pdb))

    return

def create_alpha_carbon_backbone(pdbin, pdbout):
    """Takes a pdb files and removes eveything except for the alpha carbons"""

    if not pdbout.endswith('.pdb'):
        pdbout = pdbout+'.pdb'
    if not os.path.exists(pdbin):
        raise IOError('pdbin does not exist! {!s}'.format(pdbin))
    if os.path.exists(pdbout):
        raise Exception('pdbout already exists! {!s}'.format(pdbout))

    # Initialise Commander
    PDBCUR = CommandManager('pdbcur')
    # Set Command Arguments
    PDBCUR.add_command_line_arguments('XYZIN',pdbin,'XYZOUT',pdbout)
    # Set inputs
    PDBCUR.add_standard_input(['lvatom "CA[C]:*"','END'])
    # run!
    PDBCUR.run()

    if not os.path.exists(pdbout):
        raise ExternalProgramError('PDBCUR has failed to create carbon backbone. {!s}\nOUT: {!s}\nERR: {!s}'.format(pdbin, PDBCUR.out, PDBCUR.err))

    return pdbout

def create_cryst_line(pdbin, pdbout, sg, cell):
    """Adds a cryst line to pdbin"""

    if not pdbout.endswith('.pdb'):
        pdbout = pdbout+'.pdb'
    if not os.path.exists(pdbin):
        raise IOError('pdbin does not exist! {!s}'.format(pdbin))
    if os.path.exists(pdbout):
        raise Exception('pdbout already exists! {!s}'.format(pdbout))

    # Initialise Commander
    PDBSET = CommandManager('pdbset')
    # Set Command Arguments
    PDBSET.add_command_line_arguments('XYZIN',os.path.abspath(pdbin),'XYZOUT',os.path.abspath(pdbout))
    # Set Stdin
    PDBSET.add_standard_input(['SPACEGROUP {!s}'.format(sg),'CELL {!s}'.format(' '.join(map(str,cell)))])
    # run!
    PDBSET.run()

    if not os.path.exists(pdbout):
        raise ExternalProgramError('PDBSET has failed to create cryst line for {!s}\nOUT: {!s}\nERR: {!s}'.format(pdbin, PDBSET.out, PDBSET.err))

    return PDBSET



import os, sys

import scipy.cluster

from bamboo.common import Info
from giant.stats.cluster import find_connected_groups, generate_group_idxs

from scitbx.array_family import flex

class PointCluster(object):
    def __init__(self, id, points, values):
        """Class to hold information about an identified cluster of points, with associated values"""

        # Check that both are lists for the moment
        points = list(points)
        values = list(values)
        assert len(points) == len(values)

        self.parent = None
        self.id = id
        # Store as flex arrays
        self.points = flex.vec3_double(points)
        self.values = flex.double(values)

        stats = self.values.min_max_mean()
        self.min = stats.min
        self.max = stats.max
        self.mean = stats.mean
        self.size = len(self.values)

        self.peak = self.points[values.index(max(values))]
        self.centroid = self.points.mean()

    def summary(self):
        out = []
        out.append('Point Cluster {!s} Summary'.format(self.id))
        out.append('{!s} Points'.format(self.size))
        out.append('Min:'.format(self.min))
        out.append('Max:'.format(self.max))
        out.append('Mean:'.format(self.mean))
        return '\n'.join(out)

class Event(object):
    _attributes = ['estimated_pseudo_occupancy', 'estimated_bdc']
    def __init__(self, id, cluster, info=None):
        """Class to hold information about an event in a dataset"""
        # Process cluster object (points and values for Event)
        assert isinstance(cluster, PointCluster), 'cluster must be of type PointCluster'
        self.cluster = cluster
        self.cluster.parent = self
        # Give it a name
        if id: self.id = id
        else:  self.id = cluster.id
        # Allow a parent
        self.parent = None
        # Add Meta to the object
        if info:
            assert isinstance(info, Info)
            for a in self._attributes: assert hasattr(info, a)
            self.info = info
        else:
            self.info = Info(self._attributes)

    def summary(self):
        out = []
        out.append('Event {!s} Summary'.format(self.id))
        out.append(self.cluster.summary())
        out.append(self.info.summary())
        return '\n'.join(out)

class Site(object):
    _attributes = [ 'centroid', 'num_events', 'approx_size',
                    'nearest_residues', 'near_crystal_contacts']
    def __init__(self, events=None, id=None, info=None):
        """Class to hold information about an identified site (collection of events)"""
        # List of Events
        self.children = []
        self.parent = None
        self.id = id
        # Add Meta to the object
        if info:
            assert isinstance(info, Info)
            for a in self._attributes: assert hasattr(info, a)
            self.info = info
        else:
            self.info = Info(self._attributes)

        # Add Events
        if events: self.add_events(events=events, update=True)

    def add_events(self, events, update=True):
        if isinstance(events, list):
            assert isinstance(events[0], Event), 'Added events must be of class Event'
            self.children.extend(events)
        else:
            assert isinstance(events, Event), 'Added events must be of class Event'
            self.children.append(events)
        if update: self.update()
        return self

    def apply_parentage(self):
        """Set the site as the parents of the events"""
        for e in self.children: e.parent = self
        return self

    def update(self):
        centroids = flex.vec3_double([e.cluster.centroid for e in self.children])
        self.info.centroid = centroids.mean()
        self.info.approx_size = (centroids.min(), centroids.max())
        self.info.num_events = len(self.children)
        return self

    def summary(self):
        out = []
        out.append('Site Summary')
        out.append('{!s} Events'.format(self.info.num_events))
        out.append('Centroid: {!s}'.format(self.info.centroid))
        out.append('Approx Range: {!s}'.format(self.info.approx_size))
        for s in self.children:
            out.append('-----------------------')
            out.append(s.summary())
        return '\n'.join(out)

    def sort(self, key, reverse=True):
        self.children.sort(key=key, reverse=reverse)
        return self

    def find_protein_context(self, hierarchy):
        return

class SiteList(object):
    def __init__(self, sites=None):
        """Class to hold information about multiple sites on a protein"""
        # List of Sites
        self.children = []
        self.parent = None
        self.id = None
        # Add Sites
        if sites: self.add_sites(sites)

    def add_sites(self, sites, update=True):
        if isinstance(sites, list):
            assert isinstance(sites[0], Site), 'Added sites must be of class Site'
            self.children.extend(sites)
        else:
            assert isinstance(sites, Site), 'Added sites must be of class Site'
            self.children.append(sites)
        if update:
            self.update()
        return self

    def update(self):
        # Number of events at the site
        self.num_sites = len(self.children)
        return self

    def summary(self):
        out = []
        out.append('SiteList Summary')
        out.append('{!s} Sites'.format(self.num_sites))
        for s in self.children:
            out.append('=====================================')
            out.append(s.summary())
        return '\n'.join(out)

    def sort(self, key, reverse=True):
        self.children.sort(key=key, reverse=True)
        return self

    def renumber(self, start_at=1):
        for i_c, c in enumerate(self.children): c.id=i_c+start_at
        return self

def cluster_events(events, cutoff=10, linkage='average'):

    if len(events) == 1:
        return SiteList([Site(events=events, id=1).apply_parentage()])

    centroids = [e.cluster.centroid for e in events]
    cluster_ids = scipy.cluster.hierarchy.fclusterdata( X = centroids,
                                                        t = cutoff,
                                                        criterion = 'distance',
                                                        metric    = 'euclidean',
                                                        method    = linkage )
    cluster_ids = list(cluster_ids)

    sites = []
    for s_idx, e_idxs in generate_group_idxs(cluster_ids):
        assert s_idx > 0
        new_site = Site([events[i] for i in e_idxs], id=s_idx).apply_parentage()
        sites.append(new_site)

    return SiteList(sites)

def event_distance_centroid(event1, event2):
    return (flex.double(event1.cluster.centroid) - flex.double(event2.cluster.centroid)).norm()

def event_distance_min_cluster_distance(event1, event2):
    return min([min((event1.cluster.points-x).norms()) for x in event2.cluster.points])




import numpy
import iotbx.mtz, iotbx.pdb
from bamboo.common import Info

class CrystalSummary(object):

    def __init__(self, id=None):
        self.id = id
        # File-specific
        self.mtz_file = None
        self._mtz_object = None
        self.pdb_file = None
        self._pdb_input = None
        # Crystal-specific
        self.space_group = None
        self.unit_cell = None
        self.symmetry = None
        # MTZ-Specific
        self.column_labels = None
        self.high_res = None
        self.low_res = None
        # PDB-Specific

    def mtz_object(self):
        if self._mtz_object: return self._mtz_object
        else: return iotbx.mtz.object(self.mtz_file)

    def pdb_input(self):
        if self._pdb_input: return self._pdb_input
        else: return iotbx.pdb.input(self.pdb_file)

    ######################################################
    @classmethod
    def from_mtz(cls, mtz_file=None, mtz_object=None, id=None):
        assert [mtz_file, mtz_object].count(None)==1,'Provide mtz_file OR mtz_object'
        # Initialise class (populate manually)
        new_cls = cls(id=id)
        # Store filename
        new_cls.mtz_file = mtz_file
        new_cls._mtz_object = mtz_object
        # Create mtz object of reflection data
        refl_data = new_cls.mtz_object()
        # Extract the resolution limits
        new_cls.low_res, new_cls.high_res = refl_data.max_min_resolution()
        new_cls.space_group = refl_data.space_group()
        # Crystal Properties
        crystal = refl_data.crystals()[0]
        new_cls.unit_cell = crystal.unit_cell()
        new_cls.symmetry = crystal.crystal_symmetry()
        # Column information
        new_cls.column_labels = refl_data.column_labels()
        return new_cls

    @classmethod
    def from_pdb(cls, pdb_file=None, pdb_input=None, id=None):
        assert [pdb_file, pdb_input].count(None)==1,'Provide pdb_file OR pdb_input'
        # Initialise class (populate manually)
        new_cls = cls(id=id)
        # Store filename
        new_cls.pdb_file = pdb_file
        new_cls._pdb_input = pdb_input
        # Create pdb input of structure
        pdb_input = new_cls.pdb_input()
        # Crystal Properties
        new_cls.symmetry = pdb_input.crystal_symmetry()
        new_cls.space_group = new_cls.symmetry.space_group()
        new_cls.unit_cell = new_cls.symmetry.unit_cell()
        return new_cls

class UnitCellVariation(Info):
    def __init__(self, unit_cells):
        unit_cell_params = numpy.array([uc.parameters() for uc in unit_cells])
        self.mean  = tuple(unit_cell_params.mean(axis=0))
        self.std   = tuple(unit_cell_params.std(axis=0))
        self.max   = tuple(unit_cell_params.max(axis=0))
        self.min   = tuple(unit_cell_params.min(axis=0))
        self.range = tuple(numpy.array(self.max)-numpy.array(self.min))
        self.std_perc = tuple(100.0*numpy.array(self.std)/numpy.array(self.mean))
        self.max_perc = tuple(100.0*numpy.array(self.min)/numpy.array(self.mean))
        self.min_perc = tuple(100.0*numpy.array(self.max)/numpy.array(self.mean))
        self._initialized = True
    def as_pandas_table(self):
        """Return the variation as a pandas table"""
        import pandas
        pd = pandas.DataFrame(index=['a','b','c','alpha','beta','gamma'])
        pd['min']   = self.min
        pd['mean']  = self.mean
        pd['max']   = self.max
        pd['std']   = self.std
        pd['range'] = self.range
        pd['%std']  = self.std_perc
        pd['%min']  = self.min_perc
        pd['%max']  = self.max_perc
        return pd




import pandas

def extract_residue_group_density_scores(residue_group, edstats_scores, data_table=None, rg_label=None, column_suffix=''):
    """Extract density quality metrics for a residue group from precalculated edstats scores"""

    rg_sel = residue_group
    # Set defaults
    if rg_label is None:    rg_label = (rg_sel.unique_resnames()[0]+'-'+rg_sel.parent().id+'-'+rg_sel.resseq+rg_sel.icode).replace(' ','')
    if data_table is None:  data_table = pandas.DataFrame(index=[rg_label], column=[])
    # Check validity
    if len(rg_sel.unique_resnames()) != 1: raise Exception(rg_label+': More than one residue name associated with residue group -- cannot process')

    # Extract residue scores
    ed_scores = edstats_scores.scores[(rg_sel.unique_resnames()[0], rg_sel.parent().id, rg_sel.resseq_as_int(), rg_sel.icode)]
    # Append scores to data_table
    data_table.set_value(index=rg_label, col='RSCC'+column_suffix, value=ed_scores['CCSa'])
    data_table.set_value(index=rg_label, col='RSR' +column_suffix, value=ed_scores['Ra']  )
    data_table.set_value(index=rg_label, col='B_AV'+column_suffix, value=ed_scores['BAa'] )
    data_table.set_value(index=rg_label, col='RSZO'+column_suffix, value=ed_scores['ZOa'] )
    data_table.set_value(index=rg_label, col='RSZD'+column_suffix, value=ed_scores['ZDa'] )

    return data_table









#
# Open with the command
# ccp4mg -norestore -picture testpic.mgpic.py
#
# Save image and quit with
# ccp4mg -norestore -picture testpic.mgpic.py -R file-name.png -RO '{"size":"1000x1000"}' -quit
#

# SET THE VIEW
view = View(
    centre_xyz = {{ view['camera_centre'] }},
    orientation = {{ view['orientation'] }},
#    scale = 'auto',
    scale = 30,
    fog_enabled = True,
    fog_near = -0.0,
    fog_far = 20.0,
    slab_enabled = 1,
    slab_width = 10,
    slab_offset = 0,
    clip_cap = False,
    customClipPlanes = [],
)

{% if mol %}
# OPEN MOL
MolData(
    filename =  [   'FULLPATH',
                    '{{ mol['path'] }}',
                    '{{ mol['path'] }}'
                ],
    name = '{{ mol['name'] }}',
)

# DISPLAY MOL
MolDisp(
    selection_parameters = {
                            'select' : 'cid',
                            'cid' : 'all',
                           },
    colour_parameters = {
                            'colour_mode' : 'atomtype'
                        },
)
{% endif %}

{% if map %}
# OPEN MAP
MapData(
    filename = [    'FULLPATH',
                    '{{ map['path'] }}',
                    '{{ map['path'] }}'
                ],
    name = '{{ map['name'] }}',
    expanded = 1,
    filetype = 'MAP',
)

{% for contour in map['contour'] %}
# DISPLAY DIFF MAP
MapDisp(
    opacity = 1.0,
    style = 'surface_style_solid_chickenwire',
    radius = 7,
#    clip_mode = 'ON',
    surface_style = 'surface_solid_chickenwire',
    contour_level = {{ contour }},
    contour_scale = 'absolute',
    colour = 'blue',
)
{% endfor %}
{% endif %}

{% if diff_map %}
# OPEN MAP
MapData(
    filename = [    'FULLPATH',
                    '{{ diff_map['path'] }}',
                    '{{ diff_map['path'] }}'
                ],
    name = '{{ diff_map['name'] }}',
    expanded = 1,
    filetype = 'MAP',
)

{% if diff_map['pos-contour'] %}
{% for contour in diff_map['pos-contour'] %}
# DISPLAY DIFF MAP
MapDisp(
    opacity = 1.0,
    style = 'surface_style_solid_chickenwire',
    radius = 7,
#    clip_mode = 'ON',
    surface_style = 'surface_solid_chickenwire',
    contour_level = {{ contour }},
    contour_scale = 'absolute',
    colour = 'green',
)
{% endfor %}
{% endif %}

{% if diff_map['neg-contour'] %}
{% for contour in diff_map['neg-contour'] %}
MapDisp(
    opacity = 1.0,
    style = 'surface_style_solid_chickenwire',
    radius = 7,
#    clip_mode = 'ON',
    surface_style = 'surface_solid_chickenwire',
    contour_level = {{ diff_map['neg-contour'] }},
    contour_scale = 'absolute',
    colour = 'red',
)
{% endfor %}
{% endif %}

{% endif %}

# COLOUR SETTINGS
ColourSchemeManager(
    name = 'atomtype',
    ranges = ['*', ' C', ' O', ' N', ' S', ' H', ' P'],
    colours = ['grey', 'green', 'red', 'blue', 'yellow', 'grey', 'magenta'],
    colour_wheel_direction = 'clockwise',
)

ParamsManager(
    name = 'gui_params',
    show_axes = 0,
    background_colour = [0, 0, 0],
)




import os, sys, subprocess, threading, shlex, time
from distutils.spawn import find_executable

# Code here adapted from code by Sebastian Kelm

class CommandManager(object):
    '''
        Enables commands be run on another thread, with various options

        Based on jcollado's solution:
        http://stackoverflow.com/questions/1191374/subprocess-with-timeout/4825933#4825933

        https://gist.github.com/1306188

        Modified by Sebastian Kelm: added timedout parameter
        Modified by Nicholas Pearce: added code inspired by code from the CCP4 dispatcher project
    '''

    def __init__(self, program, cmd_line_args=None, std_inp_lines=None):
        # Name of program
        self.program = shlex.split(program)
        # Check the program exists
        if self.program and not find_executable(self.program[0]):
            raise Exception("Can't find the program '{!s}'. Is it available?".format(self.program[0]))
        # Command line args and standard input args
        if cmd_line_args:
            self.cmd_line_args = cmd_line_args
        else:
            self.cmd_line_args = []
        if std_inp_lines:
            self.std_inp_lines = std_inp_lines
        else:
            self.std_inp_lines = []
        # Objects
        self.kwargs = {}
        self.thread = None
        self.process = None
        # Process Control
        self.timeout = None
        self.timedout = False
        # Response
        self.output = ''
        self.error = ''
        # Meta
        self.runtime = None

    def set_timeout(self, timeout):
        """Set various program parameters"""
        self.timeout = timeout

    def add_command_line_arguments(self, *args):
        """Add to the CMD LINE for the program"""
        for arg in args:
            if   isinstance(arg, str):  self.cmd_line_args.extend(shlex.split(arg))
            elif isinstance(arg, list): self.add_command_line_arguments(*arg)

    def add_standard_input(self, *args):
        """Add line to the STDIN for the program"""
        for arg in args:
            if   isinstance(arg, str):  self.std_inp_lines.append(arg)
            elif isinstance(arg, list): self.add_standard_input(*arg)

    def print_settings(self):
        """Print out the current settings of the object"""
        print 'Program:\n\t', ' '.join(self.program)
        print 'Command:\n\t', '\n\t'.join(self.cmd_line_args)
        print 'Inputs:\n\t',  '\n\t'.join(self.std_inp_lines)

    def _prepare_inputs_and_outputs(self):
        """Prepare the Input Pipes"""
        if self.std_inp_lines:
            self.__stdin__ = subprocess.PIPE
            self.kwargs['stdin'] = self.__stdin__
        else:
            self.__stdin__ = None
            self.kwargs.pop('stdin',None)
        self.__stdout__ = subprocess.PIPE
        self.kwargs['stdout'] = self.__stdout__
        self.__stderr__ = subprocess.PIPE
        self.kwargs['stderr'] = self.__stderr__

    def run(self):
        """Run command with keyword arguments"""
        self.run_async()
        return self.join_async()

    def run_async(self):
        """Run Command with the Keyword Arguments and then return the thread object"""
        def target():
            self.process = subprocess.Popen(args=self.program+self.cmd_line_args, **self.kwargs)
            self.output, self.error = self.process.communicate('\n'.join(self.std_inp_lines))
        # Prepare pipes
        self._prepare_inputs_and_outputs()
        # Create the new thread
        self.thread = threading.Thread(target=target)
        # Start time
        self.time_start = time.time()
        # Start the Thread
        self.thread.start()

    def join_async(self):
        """Join the thread object returned by self.run_async"""
        # Check for timeout
        self.thread.join(self.timeout)
        # Act accordingly
        if self.thread.is_alive():
            self.process.terminate()
            self.thread.join()
            self.timedout = True
            return 666
        # End time
        self.time_end = time.time()
        # Calculate runtime
        self.runtime = self.time_end-self.time_start
        # Return process exit status
        return self.process.returncode

    def write_output(self, log_file):
        """Write output to a log file"""

        with open(log_file, 'a') as fh:
            fh.write('============================>'+'\n')
            fh.write('{} Log File'.format(self.program[0])+'\n')
            fh.write('============================>'+'\n')
            fh.write('> STDOUT'+'\n')
            fh.write('============================>'+'\n')
            fh.write(self.output+'\n')
            fh.write('============================>'+'\n')
            fh.write('> STDERR'+'\n')
            fh.write('============================>'+'\n')
            fh.write(self.error+'\n')
            fh.write('============================>'+'\n')

        print('Log file for {} written to {}'.format(self.program[0], log_file))





import os, datetime

class FileManager(object):
    def __init__(self, rootdir):
        """Creates and stores output files to be neatly created and retrieved"""

        self.topdir = rootdir
        self.output_dirs = {'root':rootdir}
        self.dir_parents = {'root':None}
        self.output_files = {}

    def add_dir(self, dir_name, dir_tag, top_dir_tag=None, create=True, exists=True):
        """Store a directory `dir_name` under `dir_tag` in directory under `top_dir_tag`"""
        # Check that it hasn't been created already
        assert dir_tag not in self.output_dirs.keys()
        # If no directory given, put in roots
        if top_dir_tag is None: top_dir_tag = 'root'
        # Record the directory's parent
        self.dir_parents[dir_tag] = top_dir_tag
        # Get the path of the parent directory
        top_dir_path = self.output_dirs[top_dir_tag]
        # Create dirname and store
        self.output_dirs[dir_tag] = os.path.join(top_dir_path, dir_name)
        # Create if it doesn't exist
        if create and (not os.path.exists(self.output_dirs[dir_tag])):
            self._make_directory_if_necessary(dir_tag)
        if exists:
            assert os.path.exists(self.output_dirs[dir_tag])

    def get_dir(self, dir_tag):
        """Retrieve a dirname by it's dir_tag"""
        assert dir_tag in self.output_dirs.keys()
        return self.output_dirs[dir_tag]

    def add_file(self, file_name, file_tag, dir_tag=None):
        """Store a filename `file_name` under `file_tag` in directory under `dir_tag`"""
        # Check that it hasn't been created already
        assert file_tag not in self.output_files.keys()
        # Check the directory that it's beeing added to exists
        if dir_tag is None: dir_tag = 'root'
        dir_name = self.output_dirs[dir_tag]
        # Create filename and store
        self.output_files[file_tag] = os.path.join(dir_name, file_name)

    def get_file(self, file_tag):
        """Retrieve a filename by it's file_tag"""
        assert file_tag in self.output_files.keys()
        return self.output_files[file_tag]

    def check_and_create_directories(self):
        """Check that all directories exist and create them where necessary"""
        for dir_tag in self.output_dirs.keys():
            self._make_directory_if_necessary(dir_tag=dir_tag)

    def _make_directory_if_necessary(self, dir_tag):
        if (dir_tag != 'root') and (not os.path.exists(self.output_dirs[self.dir_parents[dir_tag]])):
            self.make_directory_if_necessary(self.dir_parents[dir_tag])
        if not os.path.exists(self.output_dirs[dir_tag]):
            os.mkdir(self.output_dirs[dir_tag])

class FileObj(object):
    def __init__(self, file):
        """File information object"""

        self.input = file

        self.path = os.path.abspath(file)
        self.tag = None
        self.dir, self.name = os.path.split(self.path)
        self.base, self.ext = os.path.splitext(self.name)

        stats = os.lstat(file)

        self.created  = datetime.datetime.fromtimestamp(stats.st_ctime)
        self.modified = datetime.datetime.fromtimestamp(stats.st_mtime)
        self.accessed = datetime.datetime.fromtimestamp(stats.st_atime)

    def __call__(self):
        return self.path




import os, sys, glob, re

#################################
import matplotlib
matplotlib.use('Agg')
matplotlib.interactive(0)
from matplotlib import pyplot
pyplot.style.use('ggplot')
#################################

import libtbx.phil

import numpy, pandas

import iotbx.pdb

from scitbx.array_family import flex

from giant.structure.ensembles import StructureCollection

#######################################

bar = '=======================++>'

#######################################

blank_arg_prepend = {'.pdb':'pdb='}

master_phil = libtbx.phil.parse("""
input {
    pdb = None
        .help = 'Input structure files'
        .type = str
        .multiple = True
    label = file_name *folder_name
        .help = 'How should the datasets be labelled?'
        .type = choice
        .multiple = False
}
output {
    out_dir = reduced_ensemble
        .type = path
}
""")


#######################################

def run(params):

    assert params.input.pdb, 'No PDBs provided'

    # Sort input files
    params.input.pdb = sorted(params.input.pdb)

    if params.input.label == 'folder_name': labels = [os.path.basename(os.path.dirname(p)) for p in params.input.pdb]
    elif params.input.label == 'file_name': labels = [os.path.basename(p) for p in params.input.pdb]

    e = StructureCollection.from_files(params.input.pdb, labels=labels)
    e.load_all()
    e.write_all(out_dir=params.output.out_dir)

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)




import networkx as nx

def convert_edges_to_graph(edges, label_hash={}):
    """Convert a list of labels and distances to a graph"""
    g = nx.Graph()
    for lab1, lab2, edge_wt in edges: g.add_edge(label_hash.get(lab1,lab1), label_hash.get(lab2,lab2), weight=edge_wt)
    return g

def convert_tree_to_edges(cluster_node):
    """Takes a scipy ClusterNode object and returns network-like edges for the object"""
    # Top of the tree
    top = cluster_node
    # Nodes yet to process
    next_nodes = [cluster_node]
    # List of connections
    connections = []
    # Iterate through nodes and create edges to their children
    while True:
        # Get the next node
        node = next_nodes.pop(0)
        # Get attached nodes
        l = node.get_left()
        r = node.get_right()
        # Iterate through attached nodes
        for child in [l,r]:
            # Append to nodes if not leaves
            if not child.is_leaf(): next_nodes.append(child)
            # Add connections and distance
            connections.append((node.get_id(), child.get_id(), node.dist-child.dist))
        # Exit if not more nodes
        if not next_nodes: break
    return connections



import os, sys, tempfile

from rdkit.Chem import AllChem as Chem

def is_valid_smiles(string):
    """Checks if a string is a valid smile string (not smarts)"""

    string = string.strip()
    # Check for invalid characters
    invalid = [',','~',' ']
    if [1 for inv in invalid if inv in string]:
        return False
    # Check brackets
    brackets = [('(',')'),('[', ']')]
    if [1 for bra in brackets if string.count(bra[0]) != string.count(bra[1])]:
        return False
    else:
        return True

def check_smile_readable(smile):
    """Checks a smile string is interpretable by Rdkit. Returns a mol object if successful"""

    mol = Chem.MolFromSmiles(smile)
    if not mol: raise SmileCheckError('Smile not rdkit-readable: {!s}'.format(smile))
    else:       return mol

def get_canonical_smiles(smile):
    """Makes a Smile String Canonical with the RDKIT framework"""

    if is_valid_smiles(smile):
        mol = Chem.MolFromSmiles(smile)
        if mol:
            return Chem.MolToSmiles(mol, isomericSmiles=True)
        else:
            return None
    else:
        mol = Chem.MolFromSmarts(smile)
        if mol:
            return Chem.MolToSmarts(mol, isomericSmiles=True)
        else:
            return None

def create_pdb_from_smile(smile, filename):
    """Given a Smile String, creates the PDB File"""

    # NEED TO EMBED MOLECULE!!!

    # Create Mol
    mol = Chem.MolFromSmiles(smile)
    # Embed Mol
    Chem.EmbedMolecule(mol)
    # Save File
    out = Chem.MolToPDBFile(mol, filename)

    return out

def get_smiles_from_file_list(filelist):
    """Given a list of PDB Files, find the Smile Strings for the Compounds in the PDB Files"""

    smileslist = []
    # Iterate through the files
    for filename in filelist:
        # Read Smiles
        smile = get_smile_from_file(filename)
        # Append to list
        if smile:
            smileslist.append(smile)

    return smileslist

def get_smile_from_file(file):
    """Given a PDB File containing ONLY a compound, returns the Smile String of the Compound"""
    block = open(file,'r').read()
    return get_smile_from_block(block)

def get_smile_from_block(block):
    """Given a PDB Block containing ONLY a compound, returns the Smile String of the Compound"""

    try:
        # Try to get Mol from MolBlock
        mol = Chem.MolFromPDBBlock(block, removeHs=False, sanitize=True)
    except Exception as Err:
        print Err
        return None
    # Check for existence of Mol
    if not mol: return None
    # Get Smile
    smile = Chem.MolToSmiles(mol, isomericSmiles=False)

    return smile

def match_smile_to_list(search, list, assignbonds=False):
    """Find 'smile' in the list of smile strings"""

    # Filter out those that are none
    filtlist = [(smi,smi) for smi in list if (smi is not None)]
    # Handle the ones with dots
    for full in [smi for smi in list if (smi is not None) and ('.' in smi)]:
        filtlist.append((full,full))
        for part in full.split('.'):
            filtlist.append((full, part))

    for filt in ['C','N','O','F']:
        # Count the number of occurences of 'filt' in smile
        count = search.upper().count(filt)
        # Get those that match
        newfiltlist = [tup for tup in filtlist if tup[1].upper().count(filt) == count]
        # update
        filtlist = newfiltlist

    # Check cyclicity of compound (filter on whether or not cyclicity is present)
    if '1' in search:
        newfiltlist = [tup for tup in filtlist if '1' in tup[1]]
        filtlist = newfiltlist
    else:
        newfiltlist = [tup for tup in filtlist if '1' not in tup[1]]
        filtlist = newfiltlist

    if assignbonds:
        # Assign bonds from the search list
        try:
            bonds = [(tup[0],assign_bond_orders_from_template_smiles(refsmile=tup[1],rawsmile=search)) for tup in filtlist]
        except:
            return [fil[0] for fil in filtlist if fil[1]]
        # Filter out no template matches
        filtlist = [pair for pair in bonds if pair[1]]

    return [fil[0] for fil in filtlist if fil[0]]

def find_structure_matches(search, ref):
    """Finds full or part matches between 'search' smile and 'ref' smile"""

    fullmatch = None
    partmatch = None
    # Assign bonds to the search mol (from PDB file)
    search_bond = assign_bond_orders_from_template_smiles(ref, search)
    # No matches - return None
    if not search_bond:
        return fullmatch, partmatch
    # Canonicalise ref
    ref = get_canonical_smiles(ref)
    # Check for type of match
    if search_bond == ref:
        fullmatch = search_bond
    elif check_for_substructure_smile_matches(ref,search_bond):
        partmatch = search_bond

    return fullmatch, partmatch

def check_for_substructure_smile_matches(refsmile,partsmile):
    """Checks to see if CheckSmile is contained within RefSmile"""

    # Generate Full Molecule from Smile
    refmol = Chem.MolFromSmiles(refsmile)
    # Generate Part Molecule from Smart
    partmol = Chem.MolFromSmarts(partsmile)

    # Check for substructure matches
    return refmol.HasSubstructMatch(partmol)

def assign_bond_orders_from_template_smiles(refsmile, rawsmile, return_mol=False):
    """Takes a template smile (with aromaticity etc.) and applies bond orders to other smile"""

    # Create rawmol
    rawmol = check_smile_readable(rawsmile)
    # Create refmol
    if is_valid_smiles(refsmile):
        refmol = Chem.MolFromSmiles(refsmile)
    else:
        refmol = Chem.MolFromSmarts(refsmile)
    # Check for existence
    if not refmol:
        return None
    # Remove Hydrogens from reference molecule
    try:
        Chem.RemoveHs(refmol, implicitOnly=True)
    except ValueError:
        return None
    # Generated Molecule with transferred bond orders
    try:
        newmol = Chem.AssignBondOrdersFromTemplate(refmol,rawmol)
    except ValueError:
        return None

    if return_mol:
        return newmol
    else:
        # Convert Back To Smiles
        newsmiles = Chem.MolToSmiles(newmol, isomericSmiles=True)
        return newsmiles




import os, sys, copy

import numpy

import iotbx.pdb

import libtbx.phil
from libtbx.utils import Sorry, null_out
from scitbx.array_family import flex

from giant.maths.geometry import pairwise_dists, is_within
from giant.stats.cluster import find_connected_groups, generate_group_idxs

#######################################

blank_arg_prepend = {'.pdb' : 'pdb='}

master_phil = libtbx.phil.parse("""
pdb = None
    .help = 'The major conformation of the protein (normally the unbound or reference structure)'
    .type = str
lig = LIG,UNL,DRG,FRG
    .help = 'Residues to be occupancy refined (comma separated list of residue identifiers, i.e. lig=LIG or lig=LIG,UNL)'
    .type = str

phenix_occ_out = 'occ_phenix.params'
    .help = 'Output occupancy coupling parameter file for phenix (allowing refinement of the water model in superposition with the ligand)'
    .type = path
refmac_occ_out = 'occ_refmac.params'
    .help = 'Output occupancy coupling parameter file for phenix (allowing refinement of the water model in superposition with the ligand)'
    .type = path

group_dist = 5
    .type = float
    .help = 'Radius around ligands to include in grouped occupancies (must have same conformer as ligand)'

clash_dist = 2
    .type = float
    .help = 'The distance at which we consider the ligand to clash with a nearby (solvent) atom. Clashing atoms will be placed into another refinement group'

overwrite = True
    .type = bool
verbose = True
    .type = bool

""")

#######################################

def generate_phenix_occupancy_params(occupancy_groups):
    """Using pairs of sets of atom groups, generate occupancy groups for phenix"""

    occ_params_template = """refinement {{\n  refine {{\n    occupancies {{\n{!s}\n    }}\n  }}\n}}"""
    constrained_group_template = """      constrained_group {{\n{!s}\n      }}"""
    selection_template = '        selection = {!s}'
    res_sel_template = '(chain {!s} and resid {!s} and altid "{!s}")'

    constrained_groups = []
    for groups in occupancy_groups:
        all_g_strs = []
        for g in groups:
            # Selections for each group
            g_sel = sorted([res_sel_template.format(ag.parent().parent().id, ag.parent().resid(), ag.altloc) for ag in g])
            g_str = selection_template.format(' or \\\n                    '.join(g_sel))
            all_g_strs.append(g_str)
        constrained_group_str = constrained_group_template.format('\n'.join(all_g_strs))
        constrained_groups.append(constrained_group_str)
    occ_params = occ_params_template.format('\n'.join(constrained_groups))
    return occ_params

def generate_refmac_occupancy_params(occupancy_groups):
    """Using pairs of sets of atom groups, generate occupancy groups for refmac"""

    selection_template = 'occupancy group id {!s} chain {!s} residue {!s} alt {!s}'
    exclude_template = 'occupancy group alts incomplete {!s}'
    final_line = 'occupancy refine'

    out_lines = []
    g_idx = 1
    for group in occupancy_groups:
        # Selections for each group
        for i_g, g in enumerate(group):
            out_lines.extend(sorted([selection_template.format(g_idx+i_g, ag.parent().parent().id, ag.parent().resseq, ag.altloc) for ag in g]))
        out_lines.append(exclude_template.format(' '.join(map(str,range(g_idx, g_idx+len(group))))))
        # Increment the group number
        g_idx += len(group)

    out_lines.append(final_line)
    occ_params = '\n'.join(out_lines)
    return occ_params

#######################################

def run(params):

    ######################################################################
    # VALIDATE INPUT
    ######################################################################

    assert params.pdb, 'No PDB File Provided'
    assert params.lig, 'No Ligand Identifier Provided (e.g. LIG or LIG,UNL)'

#    assert params.phenix_occ_out or params.refmac_occ_out, 'Must specify at least one of refmac or phenix output files'

    if params.phenix_occ_out and os.path.exists(params.phenix_occ_out):
        if params.overwrite: os.remove(params.phenix_occ_out)
        else: raise Exception('File already exists: {}'.format(params.phenix_occ_out))
    if params.refmac_occ_out and os.path.exists(params.refmac_occ_out):
        if params.overwrite: os.remove(params.refmac_occ_out)
        else: raise Exception('File already exists: {}'.format(params.refmac_occ_out))

    ######################################################################
    # READ IN INPUT FILES
    ######################################################################

    # Read in the ligand file and set each residue to the requested conformer
    pdb_obj = iotbx.pdb.hierarchy.input(params.pdb)

    ######################################################################
    # Iterate through and create refinement groups
    ######################################################################

    # Ligand resname identifiers
    lig_resns = params.lig.split(',')
    if params.verbose: print 'Looking for ligands with resname {!s}'.format(' or '.join(lig_resns))

    # Residues matching the resn that we're looking for
    matching_resn = [ag for ag in pdb_obj.hierarchy.atom_groups() if (ag.resname in lig_resns) and ag.altloc]
    if not matching_resn: raise Sorry('No Matching Ligands found')
    # Allow us to exclude resn to avoid using twice
    use_resn = [True]*len(matching_resn)

    occupancy_groups = []

    for i_lig_ag, lig_ag in enumerate(matching_resn):

        # If excluded - skip (has probably been used already)
        if not use_resn[i_lig_ag]: continue

        if params.verbose:
            print '============================================>'
            print 'Creating Occupancy Group:', lig_ag.id_str(), lig_ag.altloc

        ######################################################################
        # SELECT RESIDUES OF THE SAME CONFORMER AROUND THE LIGAND
        ######################################################################

        # Find residues with the same conformer
        altloc_ag = [lig_ag] + [ag for ag in pdb_obj.hierarchy.atom_groups() if (lig_ag.altloc == ag.altloc) and (lig_ag.id_str() != ag.id_str())]
        # Cluster the residues and chose those near the residue of interest
        if len(altloc_ag) > 1:
            # Calculate the distances between these residues
            connection_matrix = numpy.array([[(pairwise_dists(ag1.atoms().extract_xyz(), ag2.atoms().extract_xyz()).min() < params.group_dist) for ag2 in altloc_ag] for ag1 in altloc_ag])
            # Clusters these residues by cutoff distance params.group_dist
            clusters = find_connected_groups(connection_matrix=connection_matrix)

            # Select the cluster containing the ligand (the first residue)
            n_lig_clust = clusters[0]
            # Select the atom groups in this group for the selection
            ligand_group_ag = [ag for i_ag, ag in enumerate(altloc_ag) if clusters[i_ag] == n_lig_clust]
        else:
            ligand_group_ag = altloc_ag

        ######################################################################
        # FIND RESIDUES THAT CLASH WITH THE GROUP
        ######################################################################

        # Find atom groups with different conformers to the residue of interest
        diff_altloc_ag = [ag for ag in pdb_obj.hierarchy.atom_groups() if (ag.altloc) and (lig_ag.altloc != ag.altloc)]
        # Find atom groups that clash with the residue of interest
        clash_ag = [ag for ag in diff_altloc_ag if [1 for gr_ag in ligand_group_ag if is_within(params.clash_dist, gr_ag.atoms().extract_xyz(), ag.atoms().extract_xyz())]]

        # Check to see if any of the ags in matching_resn have been found and remove them
        if params.verbose: print '============================================>'
        for ag in ligand_group_ag+clash_ag:
            if ag in matching_resn:
                if params.verbose: print 'Creating Occupancy Group for:', ag.id_str(), ag.altloc
                use_resn[matching_resn.index(ag)] = False

        ######################################################################
        # GROUP ALL OF THESE RESIDUES INTO OCCUPANCY GROUPS
        ######################################################################

        # Expand atom groups to full residue groups
        all_rgs = set([ag.parent() for ag in ligand_group_ag+clash_ag])

        # Extract all atom groups from these residue groups
        all_ags = []; [all_ags.extend(rg.atom_groups()) for rg in all_rgs]

        # Pull out the matching resn ags again
        matching_ags = [ag for ag in all_ags if (ag.resname in lig_resns)]
        matching_alt = set([ag.altloc for ag in matching_ags])
        print 'Looking for residues with altlocs: {}'.format(','.join(sorted(matching_alt)))

        # Filter out residue groups
        filtered_rgs = [rg for rg in all_rgs if (len([ag.altloc for ag in rg.atom_groups()]) == 1) or matching_alt.intersection([ag.altloc for ag in rg.atom_groups()])]
        filtered_ags = []; [filtered_ags.extend(rg.atom_groups()) for rg in filtered_rgs]
        # Remove blank altlocs
        filtered_ags = [ag for ag in filtered_ags if ag.altloc]
        print '{} residue groups, {} atom groups'.format(len(filtered_rgs), len(filtered_ags))

        filtered_alt = [ag.altloc for ag in filtered_ags]
        groups = []
        for a, g in generate_group_idxs(filtered_alt):
            groups.append([filtered_ags[i] for i in g])

        ######################################################################
        # APPEND TO OUTPUT
        ######################################################################

        occupancy_groups.append(groups)

    ######################################################################
    # GENERATE OCCUPANCY RESTRAINTS FOR REFINEMENT
    ######################################################################

    ##########################################
    # REFMAC
    ##########################################
    if params.verbose:
        print '============================================>'
        print 'CREATING REFMAC OCCUPANCY REFINEMENT PARAMETERS'
    occ_params = generate_refmac_occupancy_params(occupancy_groups=occupancy_groups)
    if params.verbose:
        print '============================================>'
        print 'REFMAC Occupancy Refinement Parameter File Output'
        print '============================================>'
        print occ_params
    if params.refmac_occ_out:
        with open(params.refmac_occ_out, 'w') as fh: fh.write(occ_params)

    ##########################################
    # PHENIX PARAMETERS
    ##########################################
    if params.verbose:
        print '============================================>'
        print 'CREATING PHENIX OCCUPANCY REFINEMENT PARAMETERS'
    occ_params = generate_phenix_occupancy_params(occupancy_groups=occupancy_groups)
    if params.verbose:
        print '============================================>'
        print 'PHENIX Occupancy Refinement Parameter File Output'
        print '============================================>'
        print occ_params
    if params.phenix_occ_out:
        with open(params.phenix_occ_out, 'w') as fh: fh.write(occ_params)

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os,sys

from bamboo.maps.utils import select_factors_and_phases_for_map
from bamboo.utils.mtz import get_mtz_summary_dict
from bamboo.wrappers.maputils import fft_mtz_to_map

def convert_mtz_to_map(mtzfile, mapfile=None, method='fft', maptype='2FOFC', force=False):
    """Converts an MTZ Format File to a MAP File (using fft as default)"""

    # Create mapfile if not given
    if not mapfile:
        mapfile = os.path.splitext(mtzfile)[0]+'.'+maptype+'.ccp4'
    # Check for file validity
    if os.path.exists(mapfile):
        if force:
            os.remove(mapfile)
        else:
            return None

    # Read the Column Headings from the MTZ File
    cols = get_mtz_summary_dict(mtzfile)['colheadings']
    selcols = select_factors_and_phases_for_map(cols,maptype)

    if method == 'fft':
        writer = fft_mtz_to_map(mtzfile, mapfile, selcols)
    else:
        raise Exception('Program not recognised: {!s}'.format(method))

    return writer




settings_phil = """
settings
    .help = "General Settings"
{
    cpus = 1
        .type = int
    verbose = True
        .type = bool
    plot_graphs = True
        .help = "Output graphs using matplotlib"
        .type = bool
}
"""

image_phil = """
image
    .help = "Settings for image generation"
{
    size = 300x200
        .help = "image size of width x height (inches)"
        .type = str
}
"""



import os, sys, copy, glob
import shutil

import libtbx.phil

import numpy

from pandda.jiffies import transform_coordinates
from pandda.constants import PanddaDatasetFilenames
from giant.jiffies import merge_conformations, create_occupancy_params

############################################################################

blank_arg_prepend = {None:'select_dir='}

master_phil = libtbx.phil.parse("""
input {
    pandda_dir = ./pandda
        .help = 'Path to the pandda directory to export files from'
        .type = str
    select_dir = None
        .help = 'Manually specify directories to export'
        .type = str
        .multiple = True
}
output {
    export_dir = ./pandda-export
        .type = str

    dir_prefix = ''
        .help = 'Prefix to be added to each output (sub)directory'
        .type = str

    file_prefix = ''
        .help = 'Prefix to be added to each output file'
        .type = str

    generate_occupancy_groupings = False
        .type = bool
}
export {
    datasets_to_export = *interesting_datasets processed_datasets
        .type = choice

    files_to_export = None
        .type = path
        .multiple = True

    export_defaults = True
        .type = bool

    export_ligands = True
        .type = bool

    required_file_for_export = *model None
        .help = 'only export folder if this file exists (set to None to turn off)'
        .type = choice
}
process_and_export {
    merge_and_export_fitted_structures = True
        .help = 'export fitted ligand structures'
        .type = bool

    transform_and_export_defaults = True
        .type = bool

    pdbs_to_transform_and_export = None
        .type = path
        .multiple = True
}
templates {
    temp_prefix = 'temp-'
        .help = 'Prefix to append to temporary files'
        .type = str
    refmac_refinement = 'occ_refmac.params'
        .help = 'refmac occupancy groupings/refinement parameters'
        .type = str
    phenix_refinement = 'occ_phenix.params'
        .help = 'phenix occupancy groupings/refinement parameters'
        .type = str
}

overwrite = True
    .type = bool
verbose = False
    .type = bool

""")

############################################################################

def prepend_prefix_to_basename(prefix, path):
    return os.path.join(os.path.dirname(path),prefix+os.path.basename(path))

def set_defaults(params):

    # Add default export files
    if params.export.export_defaults:
        # Add default files to list
        params.export.files_to_export.append('*-pandda-input.pdb')
        params.export.files_to_export.append('*-pandda-input.mtz')
        params.export.files_to_export.append('*.native.ccp4')
    # Export ligands files?
    if params.export.export_ligands:
        params.export.files_to_export.append('ligand_files/*.cif')
        params.export.files_to_export.append('ligand_files/*.pdb')

    # Transform and Export default files?
    if params.process_and_export.transform_and_export_defaults:
#        params.process_and_export.pdbs_to_transform_and_export.append(PanddaDatasetFilenames.aligned_structure)
        params.process_and_export.pdbs_to_transform_and_export.append(os.path.join('modelled_structures',PanddaDatasetFilenames.modelled_structure))
    # Merge, Transform and Export fitted Structures?
    if params.process_and_export.merge_and_export_fitted_structures:
        params.process_and_export.pdbs_to_transform_and_export.append(PanddaDatasetFilenames.ensemble_structure)

    # Export occupancy groupings + refinement parameter files?
    if params.output.generate_occupancy_groupings:
        params.export.files_to_export.append(params.templates.refmac_refinement)
        params.export.files_to_export.append(params.templates.phenix_refinement)
    # Add files to be transformed to those to be exported
    for f in params.process_and_export.pdbs_to_transform_and_export:
        params.export.files_to_export.append(prepend_prefix_to_basename(params.templates.temp_prefix, f).format('*'))

def export_folder(dir, params):
    """Export a subset of a folders contents"""

    # Extract folder name
    dir_name = os.path.basename(dir)
    if params.verbose: print 'Processing Folder: {!s}'.format(dir_name)
    # Create output dir
    export_dir = os.path.join(params.output.export_dir, params.output.dir_prefix+dir_name)
    if not os.path.exists(export_dir):  os.mkdir(export_dir)
    # Export files
    if params.verbose: print 'Exporting files from {!s} to {!s}'.format(dir, export_dir)
    for proc_file_template in params.export.files_to_export:
        proc_files = glob.glob(os.path.join(dir, proc_file_template))
        for proc_file in proc_files:
            # Check that the file exists
            if not os.path.exists(proc_file):
                if params.verbose: print 'FILE DOES NOT EXIST: {!s}'.format(proc_file)
                continue
            # Exported file path
            export_file = os.path.join(export_dir, params.output.file_prefix+os.path.basename(proc_file))
            if params.verbose: print 'Exporting {!s} to {!s}'.format(proc_file, export_file)
            # Check to see if file already exists and delete if overwrite
            if os.path.exists(export_file):
                if params.overwrite: os.remove(export_file)
                else: raise Exception('File Already Exists: {}'.format(export_file))
            shutil.copy(proc_file, export_file)

def process_and_export_folder(dir, params):
    """Merge structures, transform them and export a subset of a folders contents"""

    dir_name = os.path.basename(dir)
    print '\n\n==========================================+>'
    print 'Processing Directory:', dir_name
    print '==========================================+>\n'

    ############################################################################

    # Check to see if this folder should be skipped (export fitted folders only)
    if params.export.required_file_for_export:
        if params.export.required_file_for_export == 'model':
            if not os.path.exists(os.path.join(dir, 'modelled_structures', PanddaDatasetFilenames.modelled_structure.format(dir_name))):
                print 'NO FITTED MODEL - SKIPPING: {}'.format(dir)
                return

    ############################################################################

    # MERGING WITH ALIGNED STRUCTURE

    if params.process_and_export.merge_and_export_fitted_structures:
        # Extract parameters for the merging and set them
        merging_params = merge_conformations.master_phil.extract()
        merging_params.major =  os.path.join(dir, PanddaDatasetFilenames.aligned_structure.format(dir_name))
        merging_params.minor =  os.path.join(dir, 'modelled_structures', PanddaDatasetFilenames.modelled_structure.format(dir_name))
        merging_params.output = os.path.join(dir, PanddaDatasetFilenames.ensemble_structure.format(dir_name))
        merging_params.overwrite = params.overwrite
        merging_params.verbose = params.verbose

        # Merge the structures
        print '=========================+>'
        if os.path.exists(merging_params.minor):
            print 'MERGING FITTED STRUCTURE WITH APO STRUCTURE'
            if params.verbose: print '=========================+>'
            if params.verbose: print 'MAJOR:', merging_params.major
            if params.verbose: print 'MINOR:', merging_params.minor
            if params.verbose: print 'MERGED:', merging_params.output
            merge_conformations.run(merging_params)
        else:
            print 'NOT MERGING - INPUT DOES NOT EXIST:', merging_params.minor

    ############################################################################

    print '=========================+>'
    print 'TRANSFORMING FILES FROM PANDDAS FRAME TO NATIVE FRAME'

    # Extract parameters for the transformation and set them
    transform_params = transform_coordinates.master_phil.extract()
    transform_params.dataset_pickle = os.path.join(dir, 'pickles', 'dataset.pickle')
    transform_params.overwrite = params.overwrite
    transform_params.verbose = params.verbose

    # Store files to be deleted
    temp_files_to_delete = []

    # Transform the files in the list
    for template_file in params.process_and_export.pdbs_to_transform_and_export:

        # Fill in the name of the file from the template
        dataset_file = template_file.format(dir_name)
        input_files  = glob.glob(os.path.join(dir, dataset_file))

        for f in input_files:
            transform_params.file.input = f
            transform_params.file.output = prepend_prefix_to_basename(params.templates.temp_prefix, f)

            if os.path.exists(transform_params.file.input):
                # Transform the structures
                transform_coordinates.run(transform_params)
                # Will want to delete this later
                temp_files_to_delete.append(transform_params.file.output)
            else:
                print 'NOT MAPPING - INPUT DOES NOT EXIST: {}'.format(transform_params.file.input)

    ############################################################################

    if params.output.generate_occupancy_groupings:
        print '=========================+>'
        print 'GENERATING OCCUPANCY REFINEMENT PARAMETERS'

        # Extract parameters for the occupancy parameter generation and set them
        occupancy_params = create_occupancy_params.master_phil.extract()
        occupancy_params.pdb = merging_params.output
        occupancy_params.refmac_occ_out = os.path.join(dir, params.templates.refmac_refinement)
        occupancy_params.phenix_occ_out = os.path.join(dir, params.templates.phenix_refinement)
        occupancy_params.overwrite = params.overwrite
        occupancy_params.verbose = params.verbose

        if params.verbose: print 'ANALYSING FILE: {}'.format(occupancy_params.pdb)
        if params.verbose: print 'OUTPUTTING REFMAC SETTINGS TO: {}'.format(occupancy_params.refmac_occ_out)
        if params.verbose: print 'OUTPUTTING PHENIX SETTINGS TO: {}'.format(occupancy_params.phenix_occ_out)

        try: create_occupancy_params.run(occupancy_params)
        except Exception as e:
            print 'OCCUPANCY PARAMETER GENERATION FAILED: {} - {}'.format(e, e.message)

    ############################################################################

    print '=========================+>'
    print 'EXPORTING FOLDER: {}'.format(dir)

    # Export the pandda folder
    export_folder(dir=dir, params=params)

    ############################################################################

    # Go through the files and remove the temp_prefix prefix
    print '=========================+>'
    print 'Renaming Temporary Files'
    for root, dirs, files in os.walk(os.path.join(params.output.export_dir, dir_name)):
        for f in files:
            if f.startswith(params.templates.temp_prefix):
                new_f = os.path.join(root, f[len(params.templates.temp_prefix):])
                if params.verbose: print root, f, '->', new_f
                if os.path.exists(new_f):
                    if params.overwrite: os.remove(new_f)
                    else: raise Exception('File already exists: {}'.format(new_f))
                os.rename(os.path.join(root, f), new_f)

    # Go through the temporary files and remove them
    print '=========================+>'
    print 'Deleting Temporary Files'
    for del_f in temp_files_to_delete:
        assert os.path.isfile(del_f), 'This should be a file: {!s}'.format(del_f)
        if params.verbose: print del_f, '->', '-'
        os.remove(del_f)

def run(params):

    # Default files paths
    set_defaults(params)

    # Change paths to absolute paths
    params.input.pandda_dir  = os.path.abspath(params.input.pandda_dir)
    params.output.export_dir = os.path.abspath(params.output.export_dir)

    # Must be in the pandda directory (pandda objects use relative paths)
    os.chdir(params.input.pandda_dir)

    ############################################################################

    # Find the dataset directories to be exported
    if params.input.select_dir:
        export_dirs = sorted([os.path.join(params.input.pandda_dir, params.export.datasets_to_export, p) for p in params.input.select_dir])
        export_dirs = [p for p in export_dirs if os.path.exists(p)]
        print 'Exporting:\n\t', '\n\t'.join(export_dirs)
    else:
        export_dirs = sorted(glob.glob(os.path.join(params.input.pandda_dir, params.export.datasets_to_export, '*')))
    assert export_dirs, 'No Export Directories Found'

    # Create output directory
    if not os.path.exists(params.output.export_dir): os.mkdir(params.output.export_dir)

    # Merge the fitted structures
    for dir in export_dirs:
        process_and_export_folder(dir=dir, params=params)

#######################################

if __name__ == '__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, sys, copy

import libtbx.phil

import numpy

import iotbx.pdb

from scitbx.array_family import flex

from giant.structure.utils import resolve_residue_id_clashes, merge_hierarchies
from giant.structure.occupancy import normalise_occupancies
from giant.utils.pdb import strip_pdb_to_input

############################################################################

systematic_letters = iotbx.pdb.systematic_chain_ids()

############################################################################

blank_arg_prepend = None

master_phil = libtbx.phil.parse("""
major = None
    .help = 'The major conformation of the protein (normally the unbound or reference structure)'
    .type = str

minor = None
    .help = 'The minor conformation of the protein (normally the bound or "interesting" structure)'
    .type = str

output = None
    .help = 'output pdb file'
    .type = str

overwrite = False
    .type = bool
verbose = False
    .type = bool
""")

############################################################################

def run(params):

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'VALIDATING GIVEN PARAMETERS'
    ######################################################################

    assert params.major
    assert params.minor
    if not params.output: params.output='./output.pdb'
    if os.path.exists(params.output):
        if params.overwrite: os.remove(params.output)
        else: raise Exception('File already exists: {}'.format(params.output))

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'READING INPUT FILES'
    ######################################################################

    # Read in the ligand file and set each residue to the requested conformer
    maj_obj = strip_pdb_to_input(params.major, remove_ter=True)
    min_obj = strip_pdb_to_input(params.minor, remove_ter=True)

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'VALIDATING INPUT MODELS'
    ######################################################################

    # Check that ... something
    maj_obj.hierarchy.only_model()
    min_obj.hierarchy.only_model()

    # Create a new copy of the structures
    new_major = maj_obj.hierarchy.deep_copy()
    new_minor = min_obj.hierarchy.deep_copy()

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'RESOLVING RESIDUE ID CLASHES (BY ADDING MINOR RESIDUES TO NEW CHAINS)'
    if params.verbose: print '===========================================>>>'
    ######################################################################

    new_minor = resolve_residue_id_clashes(ref_hierarchy=new_major, mov_hierarchy=new_minor)

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'MERGE THE STRUCTURES'
    ######################################################################

    final_struct = merge_hierarchies(ref_hierarchy=new_major, mov_hierarchy=new_minor)

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'NORMALISING OUTPUT STRUCTURE'
    ######################################################################

    final_struct = normalise_occupancies(hierarchy=final_struct)

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'WRITING OUTPUT STRUCTURE'
    ######################################################################

    # Update the atoms numbering
    final_struct.atoms_reset_serial()
    # Write output file
    final_struct.write_pdb_file(params.output)

    return

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



from scitbx.array_family import flex

def normalise_occupancies(hierarchy, exclude_conformers=None, max_occ=1.0, min_occ=0.0, in_place=False):
    """Normalise the occupancies of a hierarchy so that the occupancies for a residue sum to 1.0"""

    if exclude_conformers is None: exclude_conformers = []

    # Edit original copy or create new one?
    if (not in_place): hierarchy = hierarchy.deep_copy()
    # Iterate through the output structure, and normalise the occupancies if necessary
    for rg in hierarchy.residue_groups():
        # No conformers - don't normalise
        if not rg.have_conformers(): continue

        # Get the groups to change and the groups to keep constant
        ag_chnge = [ag for ag in rg.atom_groups() if ag.altloc and (ag.altloc not in exclude_conformers)]
        if not ag_chnge: continue
        ag_const = [ag for ag in rg.atom_groups() if ag.altloc and (ag.altloc in exclude_conformers)]
        # Calculate the total occupancy of the groups
        occ_chnge = sum([max(ag.atoms().extract_occ()) for ag in ag_chnge])
        occ_const = sum([max(ag.atoms().extract_occ()) for ag in ag_const])

        if (occ_const+occ_chnge < min_occ):
            # Occupancy too low - normalise to minimum value
            new_occ_chnge = min_occ - occ_const
        elif (occ_const+occ_chnge > max_occ):
            # Occupancy too high - normalise to maximum value
            new_occ_chnge = max_occ - occ_const
        else:
            # Not selected for normalising
            continue

        # Normalise residue occupancies
        assert new_occ_chnge > 0.0, 'Occupancy of INVARIABLE AGs is more than 1.0: {}'.format(occ_const)
        [ag.atoms().set_occ(ag.atoms().extract_occ()*new_occ_chnge/occ_chnge) for ag in ag_chnge]

    return hierarchy

def set_conformer_occupancy(hierarchy, conf_id, conf_occ, in_place=False):
    """Normalise the occupancies of a hierarchy so that the occupancies for a residue sum to 1.0"""

    # Edit original copy or create new one?
    if (not in_place): hierarchy = hierarchy.deep_copy()
    # Iterate through the output structure, and set the occupancies where selected
    for rg in hierarchy.residue_groups():
        # Calculate the total occupancy for this residue group
        for ag in rg.atom_groups():
            if ag.altloc == conf_id: ag.atoms().set_occ(flex.double([conf_occ]*len(ag.atoms())))
    return hierarchy




import os

import iotbx.pdb
import iotbx.mtz
import cctbx.maptbx
import scitbx.math.superpose

from bamboo.common import Meta, Info
from bamboo.common.file import FileManager
from bamboo.common.path import easy_directory

from giant.structure import make_label
from giant.structure.align import perform_flexible_alignment, find_nearest_calphas, transform_coordinates_with_flexible_alignment
from giant.xray.data import CrystalSummary
from giant.xray.symmetry import combine_hierarchies, generate_adjacent_symmetry_copies

########################################################################################################
#
#   DATASET HANDLER CLASSES
#
########################################################################################################

class DatasetHandler(object):
    def __init__(self, dataset_number, pdb_filename, mtz_filename=None, dataset_tag=None):
        """Create a dataset object to allow common functions to be applied easily to a number of datasets"""
        if pdb_filename: assert os.path.exists(pdb_filename), 'PDB file does not exist!'
        if mtz_filename: assert os.path.exists(mtz_filename), 'MTZ file does not exist!'
        # Store dataset number
        self.num = dataset_number
        # Store the tag for the dataset
        if dataset_tag:
            self.tag = dataset_tag
        else:
            # If num < 0 - mark as a reference dataset
            if self.num < 0:  self.tag = 'REF{:05d}'.format(self.num)
            else:             self.tag = 'D{:05d}'.format(self.num)
        # Output Directories
        self.output_handler = None
        self.child = None
        self.meta = Meta()
        ########################################################
        # Store filenames
        self._pdb_file = pdb_filename
        self._mtz_file = mtz_filename
        # PDB Objects
        self._structure = self.new_structure()
        # Data summaries
        if self._pdb_file: self.pdb_summary = None
        if self._mtz_file: self.mtz_summary = CrystalSummary.from_mtz(mtz_file=self._mtz_file)
        ########################################################
        # All Structure factors
        self.sfs = None
        # Truncated structure factors
        self.tr_sfs = None
        # Initialise other variables
        self.unit_cell = None
        self.space_group = None
        # Single matrix - global alignment
        self._global_rt_transform = None
        # Multiple matrices - local alignment
        self._local_rt_transforms = None
        ########################################################
        self.crystal_contact_generators = None
        ########################################################
        # Map of the clusters in the dataset
        self.events = []

    #####################################################################
    #                                                                   #
    #                       UTILITY FUNCTIONS                           #
    #                                                                   #
    #####################################################################

    def initialise_output_directory(self, outputdir):
        """Initialise a dataset output directory"""
        # Create a file and directory organiser
        self.output_handler = FileManager(rootdir=easy_directory(outputdir))

    def get_pickle_copy(self):
        """Get copy of self that can be pickled - some cctbx objects cannot be pickled..."""
        return self

    def pdb_filename(self):
        return self._pdb_file
    def mtz_filename(self):
        return self._mtz_file

    #####################################################################
    #                                                                   #
    #                       HIGH LEVEL OBJECTS                          #
    #                                                                   #
    #####################################################################

    def input(self):
        return self._structure.input
    def hierarchy(self):
        return self._structure.hierarchy

    def new_structure(self):
        """Generate a new copy of the input-hierarchy pair, from the pdb file"""
        return iotbx.pdb.hierarchy.input(file_name=self._pdb_file)

    def reflection_data(self):
        """Return an object containing the reflection data"""
        return iotbx.mtz.object(self._mtz_file)

    #####################################################################
    #                                                                   #
    #                        STRUCTURE STUFF                            #
    #                                                                   #
    #####################################################################

    def heavy_atom_sites(self):
        xray_structure = self.input().xray_structure_simple()
        return xray_structure.sites_cart().select(xray_structure.heavy_selection())
    def calpha_sites(self):
        xray_structure = self.input().xray_structure_simple()
        return xray_structure.sites_cart().select(xray_structure.backbone_selection(atom_names=['CA']))
    def backbone_sites(self):
        xray_structure = self.input().xray_structure_simple()
        return xray_structure.sites_cart().select(xray_structure.backbone_selection())

    def calphas(self):
        """Get the calphas for the structure"""
        return self.hierarchy().select(self.hierarchy().atom_selection_cache().selection('pepnames and name CA'))
    def calpha_labels(self):
        """Return the labels of the calphas of the structure"""
        return [make_label(a) for a in self.calphas().atoms_with_labels()]

    def find_nearest_calpha(self, points, hierarchy=None):
        """Returns the labels of the nearest calpha for each of the given points"""
        if hierarchy is None: hierarchy = self.hierarchy()
        return find_nearest_calphas(hierarchy, coordinates=points)

    #####################################################################
    #                                                                   #
    #                        ALIGNMENT STUFF                            #
    #                                                                   #
    #####################################################################

    def set_global_alignment(self, alignment):
        self._global_rt_transform = alignment
    def global_alignment_transform(self):
        return self._global_rt_transform

    def set_local_alignments(self, alignment):
        self._local_rt_transforms = alignment
    def local_alignment_transforms(self):
        return self._local_rt_transforms

    def transform_coordinates(self, points, method, point_mappings=None, inverse=False):
        """Transform coordinates using contained alignments"""
        assert method in ['local','global'], 'METHOD NOT DEFINED: {!s}'.format(method)
        if method == 'global':
            if inverse: return self.global_alignment_transform().inverse() * points
            else:       return self.global_alignment_transform() * points
        elif method == 'local':
            assert point_mappings is not None
            return transform_coordinates_with_flexible_alignment(   alignments  = self.local_alignment_transforms(),
                                                                    coordinates = points,
                                                                    mappings    = point_mappings,
                                                                    inverse     = inverse)

    def transform_from_reference(self, points, method, point_mappings=None):
        """Use alignment to map to reference frame from our frame"""
        return self.transform_coordinates(  points         = points,
                                            method         = method,
                                            point_mappings = point_mappings,
                                            inverse        = True   )

    def transform_to_reference(self, points, method, point_mappings=None):
        """Use alignment to map to reference frame from our frame"""
        if point_mappings is None:
            point_mappings = self.find_nearest_calpha(points)
        return self.transform_coordinates(  points         = points,
                                            method         = method,
                                            point_mappings = point_mappings,
                                            inverse        = False  )

    #####################################################################
    #                                                                   #
    #                         SYMMETRY STUFF                            #
    #                                                                   #
    #####################################################################

    def generate_symmetry_copies(self, rt_method=None, save_operators=True, buffer=10):
        """Generate the symmetry copies of the reference structure in the reference frame"""

        # Use symmetry operations to create the symmetry mates of the reference structure
        sym_ops, sym_hierarchies, chain_mappings = generate_adjacent_symmetry_copies(   ref_hierarchy    = self.new_structure().hierarchy,
                                                                                        crystal_symmetry = self.input().crystal_symmetry(),
                                                                                        buffer_thickness = buffer    )
        # Record the symmetry operations that generate the crystal contacts
        if save_operators: self.crystal_contact_generators = sym_ops

        # Create a combined hierarchy of the crystal contacts
        symmetry_root = combine_hierarchies(sym_hierarchies)
        # Transform to reference frame?
        if rt_method: symmetry_root.atoms().set_xyz(self.transform_to_reference(points=symmetry_root.atoms().extract_xyz(), method=rt_method))
        # Save coordinates
        self.symmetry_copies = symmetry_root
        return self.symmetry_copies

class ReferenceDatasetHandler(DatasetHandler):
    _origin_shift = (0,0,0)
    _binning = None

    def set_origin_shift(self, origin_shift):
        self._origin_shift = origin_shift
    def origin_shift(self):
        return self._origin_shift

    def set_map_scale(self, map_mean, map_rms):
        """Record the map mean and rms values"""
        self._map_mean = map_mean
        self._map_rms  = map_rms
    def map_scale(self):
        return (self._map_mean, self._map_rms)

########################################################################################################
#
#   MAP HANDLER CLASSES
#
########################################################################################################

def map_handler(map_data, unit_cell):
    """Map handler for easy sampling of map"""

    basic_map = cctbx.maptbx.basic_map( cctbx.maptbx.basic_map_unit_cell_flag(),
                                        map_data,
                                        map_data.focus(),
                                        unit_cell.orthogonalization_matrix(),
                                        cctbx.maptbx.out_of_bounds_clamp(0).as_handle(),
                                        unit_cell
                                    )

    return basic_map

########################################################################################################
#
#   HANDLER FUNCTIONS
#
########################################################################################################

def align_dataset_to_reference(d_handler, ref_handler, method):
    """Calculate the rotation and translation needed to align one structure to another"""
    assert method in ['both','local','global'], 'METHOD NOT DEFINED: {!s}'.format(method)
    global_rt_transform = local_rt_transforms = None
    if method == 'global' or method == 'both':
        my_sites = d_handler.calpha_sites()
        ref_sites = ref_handler.calpha_sites()
        assert len(my_sites) == len(ref_sites)
        global_rt_transform = scitbx.math.superpose.least_squares_fit(reference_sites=ref_sites, other_sites=my_sites).rt()
    if method == 'local' or method == 'both':
        local_rt_transforms = perform_flexible_alignment(mov_hierarchy=d_handler.hierarchy(), ref_hierarchy=ref_handler.hierarchy(), cutoff_radius=10)
    return global_rt_transform, local_rt_transforms





import os, sys, random

from iotbx import reflection_file_reader
from iotbx.reflection_file_utils import reflection_file_server

from cctbx.array_family import flex

def perturb_structure_factors_proportional_to_observation_error(mtzin, mtzout, seed=0):
    """Take the amplitudes from perturbation_array and use them to perturb the amplitudes in main_array (by gaussian noise)"""

    assert os.path.exists(mtzin), 'MTZ File does not exist! {!s}'.format(mtzin)
    assert not os.path.exists(mtzout), 'MTZ File already exists! {!s}'.format(mtzout)

    random.seed(seed)

    output_column_label = 'F_PERTURBED'

    # Create an mtz file object
    mtz_file = reflection_file_reader.any_reflection_file(mtzin)
    # Create file server to serve up the mtz file
    mtz_server = reflection_file_server(crystal_symmetry=None, force_symmetry=True, reflection_files=[mtz_file], err=sys.stderr)

    # At the moment hardcode a ccp4 label
    sfobs_label = 'F,SIGF'
    rfree_label = 'FreeR_flag'

    # Get the observed miller array, and the R-free flags
    sfobs_miller = mtz_server.get_miller_array(sfobs_label)
    rfree_miller = mtz_server.get_miller_array(rfree_label)

    # F column
    sfobs_amplts = sfobs_miller.data()
    # SIGF column
    sfobs_scales = sfobs_miller.sigmas()

    # Add to observed data
    perturbed_data = [f + e*random.gauss(0,1) for f,e in zip(sfobs_amplts, sfobs_scales)]
    # Zero any negative amplitudes
    perturbed_data = [f if f>0.0 else 0.0 for f in perturbed_data]

    # Generate new miller array
    perturbed_array = sfobs_miller.customized_copy(data=flex.double(perturbed_data))
    assert perturbed_array.is_real_array()

    print 'CORRELATION:', sfobs_miller.correlation(perturbed_array).coefficient()

    # Convert to mtz dataset
    perturbed_mtz_dataset = perturbed_array.as_mtz_dataset(output_column_label)
    # Copy across the R-Free
    perturbed_mtz_dataset.add_miller_array(rfree_miller, rfree_label)

    # Write out mtz file
    perturbed_mtz_object = perturbed_mtz_dataset.mtz_object()
    perturbed_mtz_object.write(file_name=mtzout)

    return output_column_label

def perturb_structure_factors_proportional_to_model_error(mtzin, mtzout, seed=0):
    """Perturb the observed structure factors proportional to the model error (FOBS-FCALC)"""

    random.seed(seed)

    assert os.path.exists(mtzin), 'MTZ File does not exist! {!s}'.format(mtzin)
    assert not os.path.exists(mtzout), 'MTZ File already exists! {!s}'.format(mtzout)

    output_column_label = 'F_PERTURBED'

    # Create an mtz file object
    mtz_file = reflection_file_reader.any_reflection_file(mtzin)
    # Create file server to serve up the mtz file
    mtz_server = reflection_file_server(crystal_symmetry=None, force_symmetry=True, reflection_files=[mtz_file], err=sys.stderr)

    # At the moment hardcode a ccp4 label
    sfobs_label = 'F,SIGF'
    sfdif_label = 'DELFWT,PHDELWT'
    rfree_label = 'FreeR_flag'

    # Get the observed miller array, and the R-free flags
    sfobs_miller = mtz_server.get_miller_array(sfobs_label)
    sfdif_miller = mtz_server.get_miller_array(sfdif_label).amplitudes()
    rfree_miller = mtz_server.get_miller_array(rfree_label)

    # F column
    sfobs_amplts = sfobs_miller.data()
    # FOBS-FCALC column
    sfobs_scales = sfdif_miller.data()

    # Add to observed data
    perturbed_data = [f + e*random.gauss(0,1) for f,e in zip(sfobs_amplts, sfobs_scales)]
    # Zero any negative amplitudes
    perturbed_data = [f if f>0.0 else 0.0 for f in perturbed_data]

    # Generate new miller array
    perturbed_array = sfobs_miller.customized_copy(data=flex.double(perturbed_data))
    assert perturbed_array.is_real_array()

    print 'CORRELATION:', sfobs_miller.correlation(perturbed_array).coefficient()

    # Convert to mtz dataset
    perturbed_mtz_dataset = perturbed_array.as_mtz_dataset(output_column_label)
    # Copy across the R-Free
    perturbed_mtz_dataset.add_miller_array(rfree_miller, rfree_label)

    # Write out mtz file
    perturbed_mtz_object = perturbed_mtz_dataset.mtz_object()
    perturbed_mtz_object.write(file_name=mtzout)

    return output_column_label

def calculate_rms_mtz(mtzfiles, mtzout):
    """Take a list of mtz files and calculate the rms of the structure factors"""

    pass






import libtbx.phil

pandda_phil = libtbx.phil.parse("""
pandda
{
    input
        .help = "File names"
    {
        data_dirs = None
            .type = str
        pdb_style = final.pdb
            .type = str
        mtz_style = None
            .type = str
        lig_style = *.cif
            .type = str
        max_new_datasets = 500
            .help = "Maximum number of new datasets that can be processed for an analysis -- used to break the analysis into chunks when processing large numbers of datasets"
            .type = int
        regex
            .help = "Advanced dataset labelling regexs"
        {
            dir_regex = None
                .type = str
            pdb_regex = None
                .type = str
            mtz_regex = None
                .type = str
        }
        filter
            .help = "Provide a dataset to filter input datasets against"
        {
            pdb = None
                .type = path
        }
        reference
            .help = "Manually define reference dataset to align+scale all other datasets to"
        {
            pdb = None
                .type = path
            mtz = None
                .type = path
            amp_label = None
                .type = str
            pha_label = None
                .type = str
        }
        flags
            .help = "Flags for individual datasets"
        {
            ignore_datasets = None
                .help = 'Reject these datasets, don\'t even load them - comma separated list of dataset tags'
                .type = str
            no_analyse = None
                .help = 'Don\'t analyse these datasets, only use them to build the distributions - comma separated list of dataset tags'
                .type = str
            no_build = None
                .help = 'Don\'t use these datasets to build the distributions, only analyse them - comma separated list of dataset tags'
                .type = str
        }
    }
    output
        .help = "Output directory"
    {
        out_dir = './pandda'
            .type = path
        new_analysis_dir = False
            .help = "Create a new analysis directory to prevent overwriting previous results"
            .type = bool
        dataset_prefix = ''
            .help = "Prefix to be attached to each dataset name"
            .type = str
        pickling
            .help = "Pickle Settings"
            .expert_level = 1
        {
            pickle_complete_pandda = False
                .help = "Pickle the entire PANDDA object for reloading - takes a lot of space"
                .type = bool
            pickle_map_analysers = False
                .help = "Pickle map analysers"
                .type = bool
            pickle_dataset_maps = False
                .help = "Pickle the dataset maps as part of the dataset pickle object"
                .type = bool
        }
        developer
            .help = "Development Settings (Not needed by most users)"
            .expert_level = 3
        {
            write_maps_for_empty_datasets = False
                .help = "Output maps for all datasets, not just datasets marked as interesting"
                .type = bool
            write_grid_masks = False
                .help = "Output the grid masks which control which areas are analysed"
                .type = bool
            write_all_z_map_types = False
                .help = "Output all possible types of Z-maps (Takes a lot of space on disk)"
                .type = bool
        }
    }
    results {
        events {
            order_by = *z_peak z_mean cluster_size
                .help = "How should events be ordered within each site?"
                .type = choice
        }
        sites {
            order_by = *num_events
                .help = "How should sites be ordered?"
                .type = choice
        }
    }
    method
        .help = "High Level control of algorithm"
    {
        reload_existing_datasets = True
            .help = "Reload existing datasets? - if False, will only load new datasets (unprocessed datasets)"
            .type = bool
        reprocess_existing_datasets = False
            .help = "Reprocess existing datasets? (Time-consuming) - if False, will only calculate z-maps for new datasets (unprocessed datasets)"
            .type = bool
        reprocess_selected_datasets = None
            .help = "Reprocess selection of datasets (comma-separated list)"
            .type = str
        recalculate_statistical_maps = False
            .help = "Recalculate all of the statistical maps? (Time-consuming) - if False, looks for existing statistical maps and uses those (reverts to True if none are found)"
            .type = bool
    }
    params
        .help = "Algorithm Parameters"
    {
        alignment
            .help = "Settings to control the alignment of the structures"
        {
            method = global *local
                .help = "How should the structures be aligned? 'global' is fast, but requires high structural conservation, whereas local is slower and accounts for structural variation"
                .type = choice
                .multiple = False
        }
        filtering
            .help = "Settings to control when datasets are rejected from the analysis"
        {
            max_rmsd_to_reference = 1.5
                .help = "Reject datasets that have a calpha rmsd of greater than this to the reference (after global alignment)"
                .type = float
                .multiple = False
            max_rfree = 0.4
                .help = 'Maximum allowed rfree for a structure (datasets above this are rejected)'
                .type = float
                .multiple = False
        }
        maps
            .help = "Settings to control how maps are generated and analysed"
        {
            ampl_label = 2FOFCWT
                .type = str
            phas_label = PH2FOFCWT
                .type = str
            scaling = none *sigma volume
                .type = choice
            resolution_factor = 0.25
                .help = 'Sampling factor for fft-ing the maps'
                .type = float
            grid_spacing = 0.5
                .help = 'Spacing of the grid points in the sampled maps (A) - fixed across resolutions'
                .type = float
            padding = 3
                .help = "Padding around the edge of the maps (A)"
                .type = float
        }
        masks
            .help = "Parameters to control the masking of grid points around the protein"
        {
            inner_mask = 1.8
                .help = "Points are masked within this distance of protein atoms"
                .type = float
            inner_mask_symmetry = 3.0
                .help = "Points are masked within this distance of neighbouring symmetry copies of protein atoms"
                .type = float
            outer_mask = 6
                .help = "Points are masked outside this distance of protein atoms"
                .type = float
        }
        analysis
            .help = "Settings to control the selection of datasets"
        {
            min_build_datasets = 40
                .help = 'Minimum number of datasets needed to build distributions'
                .type = int
            max_build_datasets = 60
                .help = 'Maximum number of datasets used to build distributions'
                .type = int
            dynamic_res_limits = True
                .help = 'Allow the analysed resolution limits to change depending on the dataset resolution ranges'
                .type = bool
            high_res_upper_limit = 0.0
                .help = 'Highest resolution limit (maps are never calulcated above this limit)'
                .type = float
            high_res_lower_limit = 4.0
                .help = 'Lowest resolution limit (datasets below this are ignored)'
                .type = float
            high_res_increment = 0.05
                .help = 'Increment of resolution shell for map analysis'
                .type = float
        }
        z_map
            .help = "Settings to control the calculation of z-maps"
            .expert_level = 3
        {
            map_type = naive adjusted uncertainty *adjusted+uncertainty
                .help = 'Type of Z-map to calculate'
                .type = choice
        }
        blob_search
            .help = "Settings to control the finding of blobs"
            .expert_level = 1
        {
            contour_level = 2.5
                .help = 'Contour level when looking for blobs'
                .type = float
            min_blob_volume = 10.0
                .help = 'Blob volume filter for detecting blobs'
                .type = float
            min_blob_z_peak = 3.0
                .help = 'Blob Z-peak filter for detecting blobs'
                .type = float
        }
        background_correction
            .help = "Parameters to control the estimation of feature background corrections"
            .expert_level = 3
        {
            min_bdc = 0.0
                .type = float
                .help = 'Minimum background correction estimate'
            max_bdc = 1.0
                .type = float
                .help = 'Maximum background correction estimate'
            increment = 0.01
                .type = float
                .help = 'Resolution of background correction estimation'
        }
    }
    exit_flags
        .help = "Flags for terminating the program early"
    {
        dry_run = False
            .help = "Find input files on disk, but exit before loading any datasets."
            .type = bool
        setup_only = False
            .help = "Load datasets and create grid partition, but exit before analysis."
            .type = bool
        calculate_first_mean_map_only = False
           .help = "Will calculate the highest resolution mean map and then exit - used for initial reference modelling."
            .type = bool
    }
}
include scope giant.phil.settings_phil
""", process_includes=True)




import os, sys
import numpy

from bamboo.common.status import status_bar

def write_numpy_array_to_file(np_array, output_file, status=True):
    """Takes numpy array and writes to file. Chooses axis of numpy array so as to maximise the number of rows of the output file"""

    assert isinstance(np_array, numpy.array), 'Array must be numpy array!'
    assert not os.path.exists(output_file), 'Output file already exists!'
    assert np_array.ndim <= 2, 'Dimension of Matrix must be less than 3!'

    # Find the longest dimension of the array
    long_num = max(np_array.shape)
    long_dim = np_array.shape.index(row_num)

    # Array is wider than it is long - need to transpose
    if long_dim == 1:
        np_array=np_array.swapaxes(0,1)

    # Now iterate through and write out the lines
    with open(output_file, 'w') as fh:
        for row in xrange(long_num):
            if status: status_bar(n=row, n_max=long_num)
            output_line = [str(round(x,3)) for x in np_array[row]]
            fh.write(', '.join(output_line)+'\n')



import os, sys, copy, glob
import shutil

import libtbx.phil

import numpy

from bamboo.common.path import delete_with_glob

############################################################################

master_phil = libtbx.phil.parse("""
input {
    pandda_dir = None
        .help = 'Path to the pandda directory to export files from'
        .type = str
}

level = *basic normal ruthless devastating
    .type = choice
    .multiple = False

skip_interesting = True
    .help = 'Clean interesting datasets as well?'
    .type = bool

verbose = False
    .type = bool

YES_I_AM_SURE = False
    .type = bool

""")

blank_arg_prepend = {None:'pandda_dir='}

help = """
Cleaning Levels:
BASIC       - Remove unnecessary files
NORMAL      - Remove easily regenerated dataset files - re-run pandda to remake: quick
RUTHLESS    - Remove almost all dataset files         - re-run pandda to remake: slow
DEVASTATING - Remove almost everything                - re-run pandda to remake: very slow

Running with DEVASTATING essentially requires the pandda to be run again from scratch.

Modelled Stuctures (in the modelled_structures folders) are NEVER deleted by pandda.clean.
Files in the "analyses" and "results_summaries" folders are NEVER deleted by pandda.clean.
Directories are also NEVER deleted (only files are deleted).
"""

############################################################################

def clean_misc(top_dir, level=0):
    print '============================================================>>>'
    print 'Cleaning Miscellaneous:'

    # Level 2 cleaning
    if level >= 2:
        delete_with_glob(os.path.join(top_dir, 'aligned_structures', '*.pdb'))

def clean_datasets(top_dir, level=0, skip_interesting=True):
    print '============================================================>>>'
    print 'Cleaning Dataset Folders:'

    for d_dir in sorted(glob.glob(os.path.join(top_dir, 'processed_datasets', '*'))):

        d_name = os.path.basename(d_dir)
        print '=======================================>>>'

        if skip_interesting and os.path.exists(os.path.join(top_dir, 'interesting_datasets', d_name)):
            print 'Not Cleaning Interesting Dataset: {}'.format(d_name)
            continue
        else:
            print 'Cleaning folder: {}'.format(d_dir)

        # Level 1 cleaning - clean maps
        if level >= 1:
            delete_with_glob(os.path.join(d_dir, '*.ccp4'))

        # Level 2 cleaning - clean structures and pickles as well
        if level >= 2:
            delete_with_glob(os.path.join(d_dir, '*.csv'))
            delete_with_glob(os.path.join(d_dir, '*.pdb'))
            delete_with_glob(os.path.join(d_dir, '*.mtz'))
            delete_with_glob(os.path.join(d_dir, '*_refine.params'))
            delete_with_glob(os.path.join(d_dir, 'pickles', '*.pickle'))
            delete_with_glob(os.path.join(d_dir, 'output_images', '*.png'))
            delete_with_glob(os.path.join(d_dir, 'ligand_files',  '*'))

    return

def clean_links(top_dir, level=0, skip_interesting=True):
    print '============================================================>>>'
    print 'Cleaning Soft Links:'

    # Level 0 cleaning
    if level >= 0:
        delete_with_glob(os.path.join(top_dir, 'resolutions', '*'))

    # Level 1 cleaning
    if level >= 1:
        pass

    # Level 2 cleaning
    if level >= 2:
        delete_with_glob(os.path.join(top_dir, 'empty_directories', '*'))
#        if not skip_interesting:
#            delete_with_glob(os.path.join(top_dir, 'interesting_datasets', '*'))

    return

def clean_statistical_maps(top_dir, level=0):
    print '============================================================>>>'
    print 'Cleaning Statistical Maps:'

    # Level 0 cleaning - clean skew and kurt maps
    if level >= 0:
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-skew_map.ccp4'))
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-kurt_map.ccp4'))
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-bimo_map.ccp4'))
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-stds_map.ccp4'))

    # Level 1 cleaning - clean mean and sadj maps
    if level >= 1:
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-mean_map.ccp4'))
        delete_with_glob(os.path.join(top_dir, 'statistical_maps', '*-sadj_map.ccp4'))

    return

def clean_pickles(top_dir, level=0):
    print '============================================================>>>'
    print 'Cleaning Pickles:'

    # Level 2 cleaning - clean meta
    if level >= 2:
        delete_with_glob(os.path.join(top_dir, 'pickled_panddas', 'dataset_meta.pickle'))
        delete_with_glob(os.path.join(top_dir, 'pickled_panddas', 'reference_dataset.pickle'))

    # Level 3 cleaning - delete everthing
    if level >= 3:
        delete_with_glob(os.path.join(top_dir, 'pickled_panddas', '*.pickle'))

def run(params):

    assert params.input.pandda_dir, 'Must specify pandda directory'

    if params.level   == 'basic':
        ferocity = 0
    elif params.level == 'normal':
        ferocity = 1
    elif params.level == 'ruthless':
        ferocity = 2
    elif params.level == 'devastating':
        ferocity = 3

    if ferocity > 1 and (not params.YES_I_AM_SURE):
        raise SystemExit('level set to {} -- are you sure? if yes, set YES_I_AM_SURE=True and re-run'.format(params.level))

    print 'Cleaning Pandda in folder: {}'.format(params.input.pandda_dir)
    print 'Running at Ferocity Level of {} ({})'.format(ferocity, params.level)

    clean_misc(
        top_dir=params.input.pandda_dir,
        level=ferocity
    )

    clean_datasets(
        top_dir=params.input.pandda_dir,
        level=ferocity,
        skip_interesting=params.skip_interesting
    )

    clean_links(
        top_dir=params.input.pandda_dir,
        level=ferocity,
        skip_interesting=params.skip_interesting
    )

    clean_statistical_maps(
        top_dir=params.input.pandda_dir,
        level=ferocity
    )

    clean_pickles(
        top_dir=params.input.pandda_dir,
        level=ferocity
    )

#######################################

if __name__ == '__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, sys, glob

#######################################

if __name__=='__main__':

    if len(sys.argv) > 1:
        proc_dirs = sorted([os.path.abspath(p) for p in sys.argv[1:] if os.path.exists(p)])
    else:
        proc_dirs = sorted([os.path.abspath(p) for p in glob.glob('./*') if os.path.isdir(p)])

    print 'Processing {} Directories'.format(len(proc_dirs))

    for d in proc_dirs:

        print 'Processing {}'.format(d)
        os.chdir(d)
        if os.path.exists('./refine.pdb'):
            continue
        else:
            os.system('giant.quick_refine params=phenix.params program=phenix *.cif *-ensemble-model.pdb *-pandda-input.mtz')




import os, sys, copy, re, glob

#################################
import matplotlib
matplotlib.use('Agg')
matplotlib.interactive(0)
from matplotlib import pyplot
pyplot.style.use('ggplot')
#################################

import libtbx.phil

import numpy, pandas

from bamboo.plot import Radar
from bamboo.edstats import Edstats

from giant.utils.pdb import strip_pdb_to_input
from giant.structure import calculate_residue_group_occupancy, calculate_residue_group_rmsd, calculate_residue_group_bfactor_ratio
from giant.xray.edstats import extract_residue_group_density_scores

#######################################

bar = '=======================++>'

#######################################

blank_arg_prepend = {'.pdb':'pdb1=', '.mtz':'mtz1='}

residue_plot_phil =  """
plot {
    remove_blank_entries = False
        .type = bool
    print_axis_values = True
        .type = bool
    parameters {
        rscc {
            title = 'Model\nQuality\n(RSCC)'
                .type = str
            axis_min = 0.60
                .type = float
            axis_max = 0.85
                .type = float
            axis_invert = True
                .type = bool
        }
        rszd {
            title = 'Model\nAccuracy\n(RSZD)'
                .type = str
            axis_min = 1.50
                .type = float
            axis_max = 4.00
                .type = float
            axis_invert = False
                .type = bool
        }
        rszo {
            title = 'Model\nPrecision\n(RSZO/OCC)'
                .type = str
            axis_min = 0.00
                .type = float
            axis_max = 2.00
                .type = float
            axis_invert = True
                .type = bool
        }
        b_factor_ratio {
            title = 'B-Factor\nRatio'
                .type = str
            axis_min = 1.00
                .type = float
            axis_max = 3.00
                .type = float
            axis_invert = False
                .type = bool
        }
        rmsd {
            title = 'Model\nRMSD'
                .type = str
            axis_min = 0.00
                .type = float
            axis_max = 1.50
                .type = float
            axis_invert = False
                .type = bool
        }
    }
}
"""

master_phil = libtbx.phil.parse("""
input {
    pdb1 = None
        .type = str
        .multiple = False
    mtz1 = None
        .type = str
        .multiple = False

    pdb2 = None
        .type = str
        .multiple = False
    mtz2 = None
        .type = str
        .multiple = False

    label = ''
        .type = str
        .multiple = False
}
selection {
    res_names = LIG,UNL,DRG
        .type = str
        .help = "Comma-separated list of residue names to score -- if None then scores all residues"
}
output {
    out_dir = ./
        .type = path
}
"""+residue_plot_phil, process_includes=True)

#######################################

def sanitise_hierarchy(hierarchy):
    hierarchy.atoms().set_chemical_element_simple_if_necessary()

def prepare_table():

    columns_fix = ['Model RMSD']
    columns_num = ['RSCC','RSZD','RSZO','OCC','RSR',
                    'Average B-factor (Residue)',
                    'Average B-factor (Surroundings)',
                    'Surroundings B-factor Ratio']
    columns_end = ['PDB','MTZ']

    columns = []
    columns.extend(columns_fix)
    columns.extend(columns_num)
    columns.extend([c+'-2' for c in columns_num])
    columns.extend(columns_end)
    columns.extend([c+'-2' for c in columns_end])

    return pandas.DataFrame(index=[], columns=columns)

def prepare_output_directory(params):
    if not os.path.exists(params.output.out_dir): os.mkdir(params.output.out_dir)
    images_dir = os.path.join(params.output.out_dir, 'residue_plots')
    if not os.path.exists(images_dir): os.mkdir(images_dir)
    return params.output.out_dir, images_dir

def score_model(params, pdb1, mtz1, pdb2=None, mtz2=None, label_prefix=''):
    """
    Score residues against density, and generate other model quality indicators.
    Identified residues in pdb1 are scored against mtz1 (and mtz2, if provided) using edstats.
    Identified residues in pdb1 are compared to the equivalent residues in pdb2, if provided.
    B-factors ratios of identified residues to surrounding sidechains are calculated.
    """

    if label_prefix: label_prefix = label_prefix + '-'

    # Extract the residues to look for
    res_names = params.selection.res_names_list

    # Extract Structure
    h1_all = strip_pdb_to_input(pdb1, remove_ter=True, remove_end=True).hierarchy
    sanitise_hierarchy(h1_all)
    h1_all = h1_all.select(h1_all.atom_selection_cache().selection('not element H'), copy_atoms=True)
    h1_pro = h1_all.select(h1_all.atom_selection_cache().selection('pepnames'), copy_atoms=True)
    h1_bck = h1_pro.select(h1_pro.atom_selection_cache().selection('(name CA or name C or name O or name N)'), copy_atoms=True)
    h1_sch = h1_pro.select(h1_pro.atom_selection_cache().selection('not (name CA or name C or name O or name N)'), copy_atoms=True)

    # Pull out residues to analyse
    if res_names: rg_for_analysis = [rg for rg in h1_all.residue_groups() if [n for n in rg.unique_resnames() if n in res_names]]
    else:         rg_for_analysis = h1_all.residue_groups()

    # Check residues to analyse or skip
    if not rg_for_analysis:
        raise Exception('There are no residues called {} in {}'.format(' or '.join(params.selection.res_names_list), pdb1))

    # Extract PDB2
    if pdb2 is not None:
        h2_all = strip_pdb_to_input(pdb2, remove_ter=True, remove_end=True).hierarchy
        sanitise_hierarchy(h2_all)
        h2_all = h2_all.select(h2_all.atom_selection_cache().selection('not element H'), copy_atoms=True)

    # Score MTZ1
    if mtz1 is not None: mtz1_edstats_scores = Edstats(mtz_file=mtz1, pdb_file=pdb1)
    else:                mtz1_edstats_scores = None
    # Score MTZ2
    if mtz2 is not None: mtz2_edstats_scores = Edstats(mtz_file=mtz2, pdb_file=pdb1)
    else:                mtz2_edstats_scores = None

    # Prepare output table
    data_table = prepare_table()

    for rg_sel in rg_for_analysis:

        # Create label for the output table
        rg_label = (label_prefix+rg_sel.unique_resnames()[0]+'-'+rg_sel.parent().id+'-'+rg_sel.resseq+rg_sel.icode).replace(' ','')

        if len(rg_sel.unique_resnames()) != 1:
            raise Exception(rg_label+': More than one residue name associated with residue group -- cannot process')

        # Append empty row to output table
        data_table = data_table.append(pandas.DataFrame(index=[rg_label]), verify_integrity=True)

        data_table.set_value(   index = rg_label, col='PDB', value=pdb1 )
        data_table.set_value(   index = rg_label,
                                col   = 'Occupancy',
                                value = calculate_residue_group_occupancy(residue_group=rg_sel) )

        data_table = calculate_residue_group_bfactor_ratio( residue_group   = rg_sel,
                                                            hierarchy       = h1_sch,
                                                            data_table      = data_table,
                                                            rg_label        = rg_label
                                                        )

        if pdb2 is not None:
            data_table.set_value(   index = rg_label, col='PDB-2', value=pdb2 )

            # Extract the equivalent residue in pdb2
            rg_sel_2 = [rg for rg in h2_all.residue_groups() if (rg.unique_resnames() == rg_sel.unique_resnames())
                                                            and (rg.parent().id       == rg_sel.parent().id)
                                                            and (rg.resseq            == rg_sel.resseq)
                                                            and (rg.icode             == rg_sel.icode)              ]
            assert rg_sel_2, 'Residue is not present in pdb file: {} not in {}'.format(rg_label, pdb2)
            assert len(rg_sel_2) == 1, 'More than one residue has been selected for {} in {}'.format(rg_label, pdb2)

            # Calculate the RMSD between the models
            try:
                rmsd = calculate_residue_group_rmsd(residue_group_1=rg_sel, residue_group_2=rg_sel_2[0])
                data_table.set_value(index=rg_label, col='Model RMSD', value=rmsd)
            except:
                pass

        # Extract Density Scores - MTZ 1
        if mtz1 is not None:
            data_table.set_value(   index = rg_label, col='MTZ', value=mtz1)
        if mtz1_edstats_scores is not None:
            data_table = extract_residue_group_density_scores(  residue_group  = rg_sel,
                                                                edstats_scores = mtz1_edstats_scores,
                                                                data_table     = data_table,
                                                                rg_label       = rg_label )
            # Normalise the RSZO by the Occupancy of the ligand
            data_table['RSZO/OCC'] = data_table['RSZO']/data_table['Occupancy']

        # Extract Density Scores - MTZ 2
        if mtz2 is not None:
            data_table.set_value(   index = rg_label, col='MTZ-2', value=mtz2)
        if mtz2_edstats_scores is not None:
            data_table = extract_residue_group_density_scores(  residue_group  = rg_sel,
                                                                edstats_scores = mtz2_edstats_scores,
                                                                data_table     = data_table,
                                                                rg_label       = rg_label,
                                                                column_suffix  = '-2' )
            # Normalise the RSZO by the Occupancy of the ligand
            data_table['RSZO/OCC-2'] = data_table['RSZO-2']/data_table['Occupancy']

    return data_table

def make_residue_radar_plot(path, data, columns, linetype=None, remove_blank_entries=False, print_axis_values=True):
    "Plot radar graph. data is a list of (label, scores) tuples. label is a string. scores is a pandas Series."

#    column_limit = [(0.6,0.85),(1.5,4),(0,2),(1,3),(0,1.5)]
#    column_names = ['RSCC','RSZD','RSZO/OCC','Surroundings B-factor Ratio','Model RMSD']
#    column_title = ['Model\nQuality\n(RSCC)', 'Model\nAccuracy\n(RSZD)', 'Model\nPrecision\n(RSZO/OCC)', 'B-Factor\nRatio', 'Model\nRMSD']
#    column_invse = [1,0,1,0,0]

    # ----------------------->
    assert isinstance(columns, dict)
    # ----------------------->
    # Column Titles are compulsory
    column_title = columns['titles']
    # ----------------------->
    # Column names are compulsory (for pulling from data_frame)
    column_names = columns['names']
    assert len(column_names) == len(column_title)
    # ----------------------->
    # Limits are optional
    if 'limits' in columns:
        column_limit = columns['limits']
        assert len(column_limit) == len(column_title)
    else:
        column_limit = None
    # ----------------------->
    # Inverse is optional
    if 'invert' in columns:
        column_invse = columns['invert']
        assert len(column_invse) == len(column_title)
    else:
        column_invse = None

    # ----------------------->
    # Extract the plot data from the data_frame
    plot_data = data[column_names]

    if remove_blank_entries:
        # ----------------------->
        # Filter the entries based on whether there is at least one values for each column
        data_mask = [True if data[c].any() else False for c in column_names]
        # ----------------------->
        # Filter against the mask
        column_title = [column_title[i] for i in range(len(data_mask)) if data_mask[i]]
        column_names = [column_names[i] for i in range(len(data_mask)) if data_mask[i]]
        if column_invse: column_invse = [column_invse[i] for i in range(len(data_mask)) if data_mask[i]]
        if column_limit: column_limit = [column_limit[i] for i in range(len(data_mask)) if data_mask[i]]
        # ----------------------->
        # Reselect the plot_data
        plot_data = data[column_names]

    # ----------------------->
    # Round column values
    col_tick_vals = []
    col_tick_labs = []
    for col in column_names:
        tvs=[]; tls=[]
        for i, v in enumerate(plot_data[col]):
            try: l = round(v, 2)
            except:
                l = 'Error'; v = None;
            # Add point marker
            tvs.append(v); tls.append(l)
        # Add to column ticks
        col_tick_vals.append(tvs); col_tick_labs.append(tls)

    # ----------------------->
    r = Radar(titles=column_title)
    # ----------------------->
    # Add each row of the data frame as a separate line
    if not linetype: linetype = ['-']*plot_data.index.size
    for label, row in plot_data.iterrows():
        r.add(row[column_names].tolist(), linetype.pop(0), label=label, lw=2)
    # ----------------------->
    # Set axis meta manually
    if column_invse: r.set_inversion(column_invse)
    if column_limit: r.set_limits(column_limit)
    # ----------------------->
    if print_axis_values: r.set_ticks(values=col_tick_vals, labels=col_tick_labs)
    else:                 r.set_ticks(values=col_tick_vals, labels=[' ']*len(col_tick_vals))
    # ----------------------->
    # Plot, modify and save
    r.plot()
    r.ax.legend(loc='lower center', fancybox=True, bbox_to_anchor=(1.0, 1.0))
    r.savefig(path)
    r.close()

    return

def format_parameters_for_plot(params):
    """Convert plot scope parameters to parameter dict"""

    p = params
    columns = {}
    columns['titles'] = [p.rscc.title,       p.rszd.title,       p.rszo.title,       p.b_factor_ratio.title,       p.rmsd.title       ]
    columns['names']  = ['RSCC','RSZD','RSZO/OCC','Surroundings B-factor Ratio','Model RMSD']
    columns['invert'] = [p.rscc.axis_invert, p.rszd.axis_invert, p.rszo.axis_invert, p.b_factor_ratio.axis_invert, p.rmsd.axis_invert ]
    columns['limits'] = [(p.rscc.axis_min,           p.rscc.axis_max),
                         (p.rszd.axis_min,           p.rszd.axis_max),
                         (p.rszo.axis_min,           p.rszo.axis_max),
                         (p.b_factor_ratio.axis_min, p.b_factor_ratio.axis_max),
                         (p.rmsd.axis_min,           p.rmsd.axis_max)  ]

    return columns

#######################################

def run(params):

    assert params.input.pdb1, 'No pdb1 provided'
    assert params.input.mtz1, 'No mtz1 provided'
    # REMOVE THIS WHEN IMPLEMENTED
    assert params.selection.res_names
    # REMOVE THIS WHEN IMPLEMENTED
    if params.selection.res_names:      params.selection.__inject__("res_names_list", params.selection.res_names.split(','))
    else:                               params.selection.__inject__("res_names_list", None)

    output_dir, images_dir = prepare_output_directory(params)
    scores_file = os.path.join(output_dir, 'residue_scores.csv')

    print bar
    print 'Scoring model...'
    data_table = score_model(   params = params,
                                pdb1   = params.input.pdb1,
                                mtz1   = params.input.mtz1,
                                pdb2   = params.input.pdb2,
                                mtz2   = params.input.mtz2,
                                label_prefix = params.input.label
                            )
    print '...Done'
    print bar

    data_table.to_csv(scores_file)
    print 'Output written to {}'.format(scores_file)
    print bar

    ###################################################################
    # Image parameters
    ###################################################################
    columns = format_parameters_for_plot(params=params.plot.parameters)

    ###################################################################
    # Output Images
    ###################################################################
    all_images = []
    print 'Generating Output Images...'
    for label, row in data_table.iterrows():
        print 'Making: {}...'.format(label)
        image_path = os.path.join(images_dir,'{}.png'.format(label))
        make_residue_radar_plot(path = image_path,
                                data = row.to_frame().T,
                                columns = columns,
                                remove_blank_entries = params.plot.remove_blank_entries,
                                print_axis_values    = params.plot.print_axis_values  )
        all_images.append(image_path)
    print '...Done.'
    print bar

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)




import itertools
import numpy, pandas

import iotbx.mtz
import scipy.cluster.hierarchy

from giant.stats.cluster import generate_group_idxs
from giant.xray.data import CrystalSummary, UnitCellVariation
from giant.xray.unit_cell import lcv_from_unit_cells, pairwise_lcv, pairwise_ecv

class CrystalGroup(object):
    def __init__(self, crystals):
        """Class to hold multiple related crystals"""
        self.crystals = crystals
        self.uc_stats = UnitCellVariation([c.unit_cell for c in self.crystals])
        self.space_groups = list(set([c.space_group.info().symbol_and_number() for c in self.crystals]))
    def size(self):
        return len(self.crystals)
    def dendrogram(self, fname, method='lcv', **kwargs):
        """Print a dendrogram of the variation in the unit cells"""
        assert method in ['ecv','lcv'], 'method not recognised'
        if   method == 'lcv': link_func = pairwise_lcv
        elif method == 'ecv': link_func = pairwise_ecv
        return unit_cell_dendrogram(fname      = fname,
                                    unit_cells = [c.unit_cell for c in self.crystals],
                                    link_func  = link_func,
                                    labels     = [c.id for c in self.crystals],
                                    **kwargs)
    ######################################################
    @classmethod
    def by_space_group(cls, crystals):
        """Group crystals by space group and return CrystalGroup for each group"""
        sg_func = lambda c: c.space_group.info().symbol_and_number()
        return [cls(list(x[1])) for x in itertools.groupby(sorted(crystals, key=sg_func), key=sg_func)]
    @classmethod
    def by_unit_cell(cls, crystals, method='lcv', cutoff=0.5):
        """Cluster crystals by unit cell and return CrystalGroup for each cluster"""
        if len(crystals) == 1: return [cls(crystals)]
        assert method in ['lcv'], 'method not recognised'
#        # Method 1
#        if   method == 'lcv': link_func = lambda a,b: lcv_from_unit_cells(a.unit_cell, b.unit_cell)
#        hierarchy = libtbx.cluster.HierarchicalClustering(crystals, link_func)
#        clusters = hierarchy.getlevel(cutoff)
#        return [cls(c) for c in clusters]
        # Method 2
        if   method == 'lcv': link_func = pairwise_lcv
        dist_mat = link_func(unit_cells=[c.unit_cell for c in crystals])
        link_mat = scipy.cluster.hierarchy.linkage(dist_mat, method='single', metric='euclidean')
        clusters = scipy.cluster.hierarchy.fcluster(link_mat, t=cutoff, criterion='distance')
        return [cls([crystals[idx] for idx in g]) for i_g,g in generate_group_idxs(clusters)]

def dendrogram(fname, link_mat, labels=None, ylab=None, xlab=None, ylim=None, annotate_y_min=0.25, num_nodes=20):
    from matplotlib import pyplot
    fig = pyplot.figure()
    ax1 = pyplot.subplot(1,1,1)
    dend = scipy.cluster.hierarchy.dendrogram(link_mat, p=num_nodes, truncate_mode='lastp', labels=labels)
    # Change labels if requested
#    if labels: ax1.set_xticklabels([labels[i] for i in dend['leaves']])
    if xlab:   ax1.set_xlabel(xlab)
    if ylab:   ax1.set_ylabel(ylab)
    if ylim:   ax1.set_ylim(ylim)
    # Make sure the labels are rotated
    xlocs, xlabels = pyplot.xticks()
    pyplot.setp(xlabels, rotation=90)
    for i, d in zip(dend['icoord'], dend['dcoord']):
        x = 0.5 * sum(i[1:3])
        y = d[1]
        if y < annotate_y_min: continue
        pyplot.plot(x, y, 'ro')
        pyplot.annotate("%.3g" % y, (x, y), xytext=(0, -8), textcoords='offset points', va='top', ha='center')
    pyplot.tight_layout()
    fig.savefig(fname)
    return fig

def unit_cell_dendrogram(fname, unit_cells, link_func, labels=None, **kwargs):
    """Calculate a dendrogram for a set of mtz files/object, clustering by unit_cell parameters"""
    dist_mat = link_func(unit_cells=unit_cells)
    link_mat = scipy.cluster.hierarchy.linkage(dist_mat, method='single', metric='euclidean')
    dendrogram(fname=fname, link_mat=link_mat, labels=labels, **kwargs)

def sort_pdbs_by_space_group(pdb_files=None, pdb_inputs=None, labels=None):
    """Group the pdbfiles by spacegroup"""
    assert [pdb_files, pdb_inputs].count(None) == 1, 'Provide pdb_files OR pdb_inputs'
    if pdb_files: pdb_inputs = [None]*len(pdb_files)
    else:         pdb_files  = [None]*len(pdb_inputs)
    # Check or create labels
    if labels: assert len(labels) == len(pdb_files), 'labels must be the same length as pdb_files/pdb_objects'
    else:      labels = range(len(pdb_files))
    # Create summary objects
    pdb_summaries = [CrystalSummary.from_pdb(pdb_file=p_f, pdb_input=p_o, id=lab) for p_f,p_o,lab in zip(pdb_files, pdb_inputs, labels)]
    # Group by SpaceGroup
    return CrystalGroup.by_space_group(pdb_summaries)

def sort_mtzs_by_spacegroup(mtz_files=None, mtz_objects=None, labels=None):
    """Group the mtzfiles by spacegroup"""
    assert [mtz_files, mtz_objects].count(None) == 1, 'Provide mtz_files OR mtz_objects'
    if mtz_files: mtz_objects = [None]*len(mtz_files)
    else:         mtz_files   = [None]*len(mtz_objects)
    # Check or create labels
    if labels: assert len(labels) == len(mtz_files), 'labels must be the same length as mtz_files/mtz_objects'
    else:      labels = range(len(mtz_files))
    # Create summary objects
    mtz_summaries = [CrystalSummary.from_mtz(mtz_file=m_f, mtz_object=m_o, id=lab) for m_f,m_o,lab in zip(mtz_files, mtz_objects, labels)]
    # Group by SpaceGroup
    return CrystalGroup.by_space_group(mtz_summaries)




import os, sys, datetime

class Log(object):
    def __init__(self, log_file=None, stdout=sys.stdout, verbose=False):
        """Log Object for writing to logs...?"""
        # File that the log will be written to
        self.log_file = log_file
        self.stdout = stdout
        self.verbose = verbose

    def __call__(self, message, show=False, hide=False):
        """Log message to file, and mirror to stdout if verbose or force_print (hide overrules show)"""

        # Format message
        if not isinstance(message, str):
            message = str(message)

        # Print to stdout
        if (not hide) and (show or self.verbose):
            self.show(message)

        if self.log_file:
            # Remove \r from message as this spoils the log (^Ms)
            message = message.replace('\r','')
            # Write to file
            self.write(message=message+'\n', mode='a')

    def show(self, message):
        print(message)

    def write(self, message, mode='a'):
        with open(self.log_file, mode) as fh: fh.write(message)

    def read_all(self):
        return open(self.log_file, 'r').read()

##################################################################
#
#   TODO Redo all code below this line - and move to utils TODO
#
##################################################################

class LigandRecord(object):

    def __str__(self):
        return 'LigandRecord: {!s}'.format(self.name)

    def __call__(self):
        pass

    def __init__(self, name='ligand', smile=None, builder=None):

        # Name of Ligand
        self.name = name
        # Program used to generate the ligand model
        self.builder = builder
        # Initialise
        self.smile = smile
        # Ligand Model (Generated from smile, random pose)
        self.unfitted = None
        self.restraints = None
        # Correctly oriented and positioned ligand file (relative to protein)
        self.fitted = None
        # Fitted ligand merged with protein structure PDB
        self.merged = None
        # Refined Merged
        self.refined = None
        # LogFiles
        self.buildinglog = None
        self.fittinglog = None
        self.refininglog = None

class ModelHistory(object):
    """Class for containing the modelling history of a Model"""

    def __init__(self, name='ModelHistory'):

        self.name = name
        self.history = []
        self.original = None

    def __getitem__(self, index):
        return self.history[index]

    def __str__(self):
        return 'ModelHistory: {!s}'.format(self.name)

    def __iter__(self):
        return iter(self.history)

    def set_original_file(self, origfile):
        """Marks a file as the first file"""
        self.original = origfile

    def add_file(self, newfile, index=0):
        """Appends a new model to the beginning of the history list"""
        self.history.insert(index, newfile)

    def get_current_file(self):
        """Returns the newest model (top of the list)"""
        return self.history[0]

class FileHistory(object):

    def __init__(self, name=None, path=None):

        # Name of object
        self.name = name
        # File history
        self.history = ModelHistory('{!s} History'.format(self.name))
        # Current Models (most recent)
        self.current = path
        self.previous = None
        self.history.set_original_file(self.current)

    def __call__(self):
        return self.current




import iotbx.ccp4_map
import cctbx.maptbx

from scitbx.array_family import flex

from pandda import analyse_graphs

def write_map_value_distribution(map_vals, output_file, plot_indices=None, plot_normal=False):
    """Write out the value distribution for a map"""
    try:
        if plot_indices: plot_vals = [map_vals[i] for i in plot_indices]
        else:            plot_vals = list(map_vals)
        analyse_graphs.map_value_distribution(f_name=output_file, plot_vals=plot_vals, plot_normal=plot_normal)
    except: pass

def write_qq_plot_against_normal(map_vals, output_file, plot_indices=None):
    """Plot the values in map_vals against those expected from a normal distribution"""
    try:
        if plot_indices: plot_vals = [map_vals[i] for i in plot_indices]
        else:            plot_vals = list(map_vals)
        analyse_graphs.qq_plot_against_normal(f_name=output_file, plot_vals=plot_vals)
    except: pass

def write_array_to_map(output_file, map_data, grid):
    """Take array on the reference grid and write to map"""
    iotbx.ccp4_map.write_ccp4_map(  file_name   = output_file,
                                    unit_cell   = grid.unit_cell(),
                                    space_group = grid.space_group(),
                                    map_data    = map_data,
                                    labels      = flex.std_string(['Map from pandda'])     )

def rotate_map(grid, d_handler, map_data, align_on_grid_point=None):
    """Apply an RT matrix to an array on the reference grid"""

    if (align_on_grid_point is not None):
        # For the local alignment transformation
        rt_lab = grid.partition().query_by_grid_points([align_on_grid_point])[0]
        print('=> Aligning Event Map to: Chain {}, Residue {}'.format(rt_lab[0], rt_lab[1].strip()))
        rt = d_handler.local_alignment_transforms()[rt_lab]
    else:
        # For the global alignment transformation
        rt = d_handler.global_alignment_transform()

    return cctbx.maptbx.rotate_translate_map(   unit_cell          = grid.unit_cell(),
                                                map_data           = map_data,
                                                rotation_matrix    = rt.r.elems,
                                                translation_vector = rt.t.elems    )

